from unittest.mock import MagicMock, patch

from ninja import NinjaAPI

from django.test import TestCase
from django.utils import timezone
from eveuniverse.models import EveMarketPrice

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from app_utils.testing import add_character_to_user, create_user_from_evecharacter

from ledger.api.ledger import LedgerApiEndpoints
from ledger.models.characteraudit import CharacterMiningLedger
from ledger.tests.test_api import _billboardcorpdata
from ledger.tests.test_api._billboardchardata import (
    CharmonthlyMarch,
    CharmonthlyMarchMulti,
    CharNoMining,
    Charyearly,
    noData,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all


class ManageApiLedgerCharEndpointsTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.char_audit_admin_manager",
                "ledger.advanced_access",
                "ledger.corp_audit_admin_manager",
            ],
        )
        cls.user2, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
            ],
        )
        cls.user3, _ = create_user_from_evecharacter(
            1003,
            permissions=[
                "ledger.basic_access",
                "ledger.advanced_access",
            ],
        )

        cls.api = NinjaAPI()
        cls.manage_api_endpoints = LedgerApiEndpoints(api=cls.api)

    def test_get_corp_ally_billbboard_api_many_users(self):
        create_user_from_evecharacter(1010)
        create_user_from_evecharacter(1011)
        create_user_from_evecharacter(1012)
        create_user_from_evecharacter(1013)
        create_user_from_evecharacter(1014)
        create_user_from_evecharacter(1015)
        create_user_from_evecharacter(1016)
        create_user_from_evecharacter(1017)
        create_user_from_evecharacter(1018)
        create_user_from_evecharacter(1019)
        create_user_from_evecharacter(1020)
        create_user_from_evecharacter(1021)
        create_user_from_evecharacter(1022)

        self.client.force_login(self.user)
        url = "/ledger/api/corporation/2001/billboard/date/2023-01-01/view/year/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

        self.client.force_login(self.user)
        url = "/ledger/api/alliance/3001/billboard/date/2023-01-01/view/year/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

    def test_get_character_billbboard_api(self):
        self.client.force_login(self.user)
        url = "/ledger/api/character/0/billboard/date/2024-03-01/view/month/"

        fixed_date = timezone.datetime(2025, 1, 31)

        with patch("django.utils.timezone.datetime") as mock_datetime:
            mock_datetime.now.return_value = fixed_date
            mock_datetime.side_effect = lambda *args, **kw: timezone.datetime(
                *args, **kw
            )

            response = self.client.get(url)

            expected_data = CharmonthlyMarch

            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.json(), expected_data)

            # Corporation
            url = "/ledger/api/corporation/0/billboard/date/2024-03-01/view/month/"

            response = self.client.get(url)

            self.assertEqual(response.status_code, 200)

            # Alliance
            url = "/ledger/api/alliance/0/billboard/date/2024-03-01/view/month/"

            response = self.client.get(url)

            self.assertEqual(response.status_code, 200)

    def test_get_character_billbboard_api_single(self):
        self.client.force_login(self.user)

        fixed_date = timezone.make_aware(timezone.datetime(2025, 1, 31))

        with patch("django.utils.timezone.datetime") as mock_datetime:
            mock_datetime.now.return_value = fixed_date
            mock_datetime.side_effect = lambda *args, **kw: timezone.make_aware(
                timezone.datetime(*args, **kw)
            )

            url = "/ledger/api/character/1001/billboard/date/2024-03-01/view/month/"
            # when
            response = self.client.get(url)
            # then
            expected_data = CharmonthlyMarch
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.json(), expected_data)

            # Corporation
            self.client.force_login(self.user)
            url = "/ledger/api/corporation/2002/billboard/date/2024-01-01/view/year/"

            response = self.client.get(url)

            self.assertEqual(response.status_code, 200)

    def test_get_billbboard_api_year(self):
        self.client.force_login(self.user)

        # Corporation
        url = "/ledger/api/corporation/0/billboard/date/2024-01-01/view/year/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

        # Alliance

        url = "/ledger/api/alliance/0/billboard/date/2024-01-01/view/year/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

    def test_get_billbboard_api_year_single(self):
        # given
        self.client.force_login(self.user)
        url = "/ledger/api/character/1001/billboard/date/2024-01-01/view/year/"
        expected_data = Charyearly

        fixed_date = timezone.datetime(2025, 1, 31)

        with patch("django.utils.timezone.datetime") as mock_datetime:
            mock_datetime.now.return_value = fixed_date
            mock_datetime.side_effect = lambda *args, **kw: timezone.datetime(
                *args, **kw
            )

            # when
            response = self.client.get(url)
            # then
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.json(), expected_data)

            # Corporation
            url = "/ledger/api/corporation/2002/billboard/date/2024-01-01/view/year/"

            response = self.client.get(url)

            self.assertEqual(response.status_code, 200)

            # Alliance
            url = "/ledger/api/alliance/3002/billboard/date/2024-01-01/view/year/"

            response = self.client.get(url)

            self.assertEqual(response.status_code, 200)

    def test_get_character_billbboard_api_multi(self):
        chars = EveCharacter.objects.filter(character_id__in=[1002, 1003])
        for char in chars:
            add_character_to_user(self.user, char)
        self.client.force_login(self.user)
        url = "/ledger/api/character/0/billboard/date/2024-03-01/view/month/"

        fixed_date = timezone.datetime(2025, 1, 31)

        with patch("django.utils.timezone.datetime") as mock_datetime:
            mock_datetime.now.return_value = fixed_date
            mock_datetime.side_effect = lambda *args, **kw: timezone.datetime(
                *args, **kw
            )

            response = self.client.get(url)

            expected_data = CharmonthlyMarchMulti
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.json(), expected_data)

    def test_get_character_billbboard_api_no_permission(self):
        self.client.force_login(self.user2)
        url = "/ledger/api/character/1001/billboard/date/2024-03-01/view/month/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)

        # Corporation
        url = "/ledger/api/corporation/2001/billboard/date/2024-03-01/view/month/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)

        # Alliance
        url = "/ledger/api/alliance/3001/billboard/date/2024-03-01/view/month/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)

    def test_get_character_billbboard_api_no_data(self):
        # given
        self.client.force_login(self.user3)

        fixed_date = timezone.datetime(2025, 1, 31)

        with patch("django.utils.timezone.datetime") as mock_datetime:
            mock_datetime.now.return_value = fixed_date
            mock_datetime.side_effect = lambda *args, **kw: timezone.datetime(
                *args, **kw
            )

            url = "/ledger/api/character/0/billboard/date/2024-03-01/view/month/"

            # when
            response = self.client.get(url)
            # then
            expected_data = noData
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.json(), expected_data)

            # Corporation
            url = "/ledger/api/corporation/0/billboard/date/2000-03-01/view/month/"

            response = self.client.get(url)

            expected_data = _billboardcorpdata.noData
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.json(), expected_data)

            # Alliance
            url = "/ledger/api/alliance/0/billboard/date/2000-03-01/view/month/"

            response = self.client.get(url)

            expected_data = _billboardcorpdata.noData
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.json(), expected_data)

    def test_get_character_billbboard_api_no_mining(self):
        # given
        self.client.force_login(self.user)
        url = "/ledger/api/character/0/billboard/date/2024-03-01/view/month/"
        CharacterMiningLedger.objects.all().delete()
        EveMarketPrice.objects.all().delete()

        fixed_date = timezone.datetime(2025, 1, 31)

        with patch("django.utils.timezone.datetime") as mock_datetime:
            mock_datetime.now.return_value = fixed_date
            mock_datetime.side_effect = lambda *args, **kw: timezone.datetime(
                *args, **kw
            )

            # when
            response = self.client.get(url)
            # then
            expected_data = CharNoMining
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.json(), expected_data)

    def test_get_corporation_billbboard_api_not_found(self):
        self.client.force_login(self.user2)

        # Corporation
        url = "/ledger/api/corporation/2001/billboard/date/2024-03-01/view/month/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), "Permission Denied")

        # Alliance

        url = "/ledger/api/alliance/3001/billboard/date/2024-03-01/view/month/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), "Permission Denied")



from collections import defaultdict

from django.db import models
from django.db.models import Case, DecimalField, F, Q, Sum, Value, When
from django.db.models.functions import Coalesce
from django.utils import timezone

from allianceauth.eveonline.models import EveCharacter

from ledger.hooks import get_extension_logger
from ledger.view_helpers.core import events_filter

logger = get_extension_logger(__name__)
# PvE - Income
BOUNTY_PRIZES = ["bounty_prizes"]
ESS_TRANSFER = ["ess_escrow_transfer"]
MISSION_REWARD = ["agent_mission_reward", "agent_mission_time_bonus_reward"]
INCURSION = ["corporate_reward_payout"]

# Cost Ref Types
CONTRACT_COST = [
    "contract_price",
    "contract_collateral",
    "contract_reward_deposited",
    "contract_brokers_fee",
    "contract_sales_tax",
]
MARKET_COST = ["market_escrow", "transaction_tax", "market_provider_tax", "brokers_fee"]
ASSETS_COST = ["asset_safety_recovery_tax"]
TRAVELING_COST = [
    "structure_gate_jump",
    "jump_clone_activation_fee",
    "jump_clone_installation_fee",
]
PRODUCTION_COST = [
    "industry_job_tax",
    "manufacturing",
    "researching_time_productivity",
    "researching_material_productivity",
    "copying",
    "reprocessing_tax",
    "reaction",
]
SKILL_COST = ["skill_purchase"]
INSURANCE_COST = ["insurance"]
PLANETARY_COST = [
    "planetary_export_tax",
    "planetary_import_tax",
    "planetary_construction",
]
LP_COST = ["lp_store"]
# Trading
MARKET_INCOME = ["market_transaction", "market_escrow"]
CONTRACT_INCOME = [
    "contract_price_payment_corp",
    "contract_reward",
    "contract_price",
    "contract_reward_refund",
    "contract_collateral_refund",
    "contract_deposit_refund",
]
DONATION_INCOME = ["player_donation"]
INSURANCE_INCOME = ["insurance"]
# MISC
MILESTONE_REWARD = ["milestone_reward_payment"]
DAILY_GOAL_REWARD = ["daily_goal_payouts"]

# PVE
BOUNTY_FILTER = Q(ref_type__in=BOUNTY_PRIZES)
MISSION_FILTER = Q(ref_type__in=MISSION_REWARD)
ESS_FILTER = Q(ref_type__in=ESS_TRANSFER)
INCURSION_FILTER = Q(ref_type__in=INCURSION)
DAILY_GOAL_REWARD_FILTER = Q(ref_type__in=DAILY_GOAL_REWARD, amount__gt=0)
# COSTS
CONTRACT_COST_FILTER = Q(ref_type__in=CONTRACT_COST, amount__lt=0)
MARKET_COST_FILTER = Q(ref_type__in=MARKET_COST, amount__lt=0)
ASSETS_COST_FILTER = Q(ref_type__in=ASSETS_COST, amount__lt=0)
TRAVELING_COST_FILTER = Q(ref_type__in=TRAVELING_COST, amount__lt=0)
PRODUCTION_COST_FILTER = Q(ref_type__in=PRODUCTION_COST, amount__lt=0)
SKILL_COST_FILTER = Q(ref_type__in=SKILL_COST, amount__lt=0)
INSURANCE_COST_FILTER = Q(ref_type__in=INSURANCE_COST, amount__lt=0)
PLANETARY_COST_FILTER = Q(ref_type__in=PLANETARY_COST, amount__lt=0)
LP_COST_FILTER = Q(ref_type__in=LP_COST, amount__lt=0)
# TRADES
MARKET_INCOME_FILTER = Q(ref_type__in=MARKET_INCOME, amount__gt=0)
CONTRACT_INCOME_FILTER = Q(ref_type__in=CONTRACT_INCOME, amount__gt=0)
DONATION_INCOME_FILTER = Q(ref_type__in=DONATION_INCOME, amount__gt=0)
INSURANCE_INCOME_FILTER = Q(ref_type__in=INSURANCE_INCOME, amount__gt=0)
MILESTONE_REWARD_FILTER = Q(ref_type__in=MILESTONE_REWARD, amount__gt=0)

PVE_FILTER = BOUNTY_FILTER | ESS_FILTER
MISC_FILTER = (
    MARKET_INCOME_FILTER
    | INSURANCE_INCOME_FILTER
    | MISSION_FILTER
    | INCURSION_FILTER
    | MILESTONE_REWARD_FILTER
)
COST_FILTER = (
    CONTRACT_COST_FILTER
    | MARKET_COST_FILTER
    | ASSETS_COST_FILTER
    | TRAVELING_COST_FILTER
    | PRODUCTION_COST_FILTER
    | SKILL_COST_FILTER
    | INSURANCE_COST_FILTER
    | PLANETARY_COST_FILTER
    | LP_COST_FILTER
)


class CharWalletIncomeFilter(models.QuerySet):
    # PvE - Income
    def annotate_bounty_income(self) -> models.QuerySet:
        return self.annotate(
            bounty_income=Coalesce(
                Sum(
                    "amount",
                    filter=BOUNTY_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_mission_income(self) -> models.QuerySet:
        return self.annotate(
            mission_income=Coalesce(
                Sum(
                    "amount",
                    filter=MISSION_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_incursion_income(self) -> models.QuerySet:
        return self.annotate(
            incursion_income=Coalesce(
                Sum(
                    "amount",
                    filter=INCURSION_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    # Trading - Income
    def annotate_market_income(self) -> models.QuerySet:
        return self.annotate(
            market_income=Coalesce(
                Sum(
                    "amount",
                    filter=MARKET_INCOME_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_contract_income(self) -> models.QuerySet:
        return self.annotate(
            contract_income=Coalesce(
                Sum(
                    "amount",
                    filter=CONTRACT_INCOME_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_donation_income(self, exclude: list = None) -> models.QuerySet:
        qs = self

        if exclude is None:
            exclude = []

        return qs.annotate(
            donation_income=Coalesce(
                Sum(
                    "amount",
                    filter=Q(DONATION_INCOME_FILTER) & ~Q(first_party_id__in=exclude),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_insurance_income(self) -> models.QuerySet:
        return self.annotate(
            insurance_income=Coalesce(
                Sum(
                    "amount",
                    filter=INSURANCE_INCOME_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_milestone_income(self) -> models.QuerySet:
        return self.annotate(
            milestone_income=Coalesce(
                Sum(
                    "amount",
                    filter=MILESTONE_REWARD_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )


class CharWalletOutSideFilter(CharWalletIncomeFilter):
    def annotate_miscellaneous(self) -> models.QuerySet:
        return self.annotate(
            miscellaneous=Coalesce(
                Sum(
                    "amount",
                    filter=MISC_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_costs(self) -> models.QuerySet:
        return self.annotate(
            costs=Coalesce(
                Sum(
                    "amount",
                    filter=COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )


class CharWalletCostQueryFilter(CharWalletOutSideFilter):
    # Costs
    def annotate_contract_cost(self) -> models.QuerySet:
        return self.annotate(
            contract_cost=Coalesce(
                Sum(
                    "amount",
                    filter=CONTRACT_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_market_cost(self) -> models.QuerySet:
        return self.annotate(
            market_cost=Coalesce(
                Sum(
                    "amount",
                    filter=MARKET_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_asset_cost(self) -> models.QuerySet:
        return self.annotate(
            asset_cost=Coalesce(
                Sum(
                    "amount",
                    filter=ASSETS_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_traveling_cost(self) -> models.QuerySet:
        return self.annotate(
            traveling_cost=Coalesce(
                Sum(
                    "amount",
                    filter=TRAVELING_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_production_cost(self) -> models.QuerySet:
        return self.annotate(
            production_cost=Coalesce(
                Sum(
                    "amount",
                    filter=PRODUCTION_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_skill_cost(self) -> models.QuerySet:
        return self.annotate(
            skill_cost=Coalesce(
                Sum(
                    "amount",
                    filter=SKILL_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_insurance_cost(self) -> models.QuerySet:
        return self.annotate(
            insurance_cost=Coalesce(
                Sum(
                    "amount",
                    filter=INSURANCE_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_planetary_cost(self) -> models.QuerySet:
        return self.annotate(
            planetary_cost=Coalesce(
                Sum(
                    "amount",
                    filter=PLANETARY_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_lp_cost(self) -> models.QuerySet:
        return self.annotate(
            lp_cost=Coalesce(
                Sum(
                    "amount",
                    filter=LP_COST_FILTER,
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )


class CharWalletQuerySet(CharWalletCostQueryFilter):
    def _get_main_and_alts(self, characters: list[EveCharacter]) -> tuple[dict, set]:
        characters_dict = {}
        char_list = []
        for char in characters:
            try:
                characters_dict[char.character_id] = char
                char_list.append(char.character_id)
            except AttributeError:
                continue
        return characters_dict, set(char_list)

    def _get_models_qs(
        self, character_ids: list, filter_date: Q
    ) -> tuple[models.QuerySet, models.QuerySet]:
        """Get the models queryset for the character ledger"""
        # pylint: disable=import-outside-toplevel
        from ledger.models.characteraudit import CharacterMiningLedger
        from ledger.models.corporationaudit import CorporationWalletJournalEntry

        # Call annotate_ledger and store the result
        char_mining_journal = CharacterMiningLedger.objects.filter(
            Q(character__character__character_id__in=character_ids) & Q(filter_date)
        )

        corp_character_journal = CorporationWalletJournalEntry.objects.filter(
            Q(second_party_id__in=character_ids) & Q(filter_date)
        )
        return char_mining_journal, corp_character_journal

    def annotate_ledger_data(self, exclude: list | None) -> models.QuerySet:
        """Get the ledger data"""
        return (
            self
            # PvE
            .annotate_bounty_income()
            # Income
            .annotate_mission_income()
            .annotate_incursion_income()
            .annotate_insurance_income()
            .annotate_market_income()
            .annotate_contract_income()
            .annotate_donation_income(exclude=exclude)
            .annotate_milestone_income()
            # Costs
            .annotate_market_cost()
            .annotate_production_cost()
            .annotate_contract_cost()
            .annotate_lp_cost()
            .annotate_traveling_cost()
            .annotate_asset_cost()
            .annotate_skill_cost()
            .annotate_insurance_cost()
            .annotate_planetary_cost()
        )

    def generate_ledger(
        self, characters: list[EveCharacter], filter_date, exclude: list | None
    ) -> tuple[models.QuerySet, models.QuerySet, models.QuerySet]:
        characters, char_list = self._get_main_and_alts(characters)

        qs = self.filter(Q(character__character__character_id__in=char_list))

        qs = qs.filter(filter_date)

        mining_qs, corp_qs = self._get_models_qs(char_list, filter_date)

        # Fiter Tax Events
        corp_qs = events_filter(corp_qs)

        characters = qs.annotate(
            char_id=Case(
                *[
                    When(
                        (Q(character__character__character_id=main_id)),
                        then=Value(char.character_id),
                    )
                    for main_id, char in characters.items()
                ],
                output_field=models.IntegerField(),
            ),
            char_name=Case(
                *[
                    When(
                        (Q(character__character__character_id=main_id)),
                        then=Value(char.character_name),
                    )
                    for main_id, char in characters.items()
                ],
                output_field=models.CharField(),
            ),
        ).values("char_id", "char_name")

        # Annotate All Ledger Data
        characters = characters.annotate_ledger_data(exclude)

        char_qs = characters.annotate(
            miscellaneous=Coalesce(
                F("mission_income")
                + F("incursion_income")
                + F("insurance_income")
                + F("market_income")
                + F("contract_income")
                + F("donation_income")
                + F("milestone_income"),
                Value(0),
                output_field=DecimalField(),
            ),
            costs=Coalesce(
                F("market_cost")
                + F("production_cost")
                + F("contract_cost")
                + F("lp_cost")
                + F("traveling_cost")
                + F("asset_cost")
                + F("skill_cost")
                + F("insurance_cost")
                + F("planetary_cost"),
                Value(0),
                output_field=DecimalField(),
            ),
        )
        return char_qs, mining_qs, corp_qs

    def generate_template(
        self,
        amounts: defaultdict,
        character_ids: list,
        filter_date: timezone.datetime,
        exclude=None,
    ) -> dict:
        """Generate data template for the ledger character information view"""
        # Define the types and their respective filters
        type_names = [
            # PvE
            "bounty_income",
            # Costs
            "market_cost",
            "production_cost",
            "contract_cost",
            "lp_cost",
            "traveling_cost",
            "asset_cost",
            "skill_cost",
            "insurance_cost",
            "planetary_cost",
            # Income
            "mission_income",
            "incursion_income",
            "insurance_income",
            "market_income",
            "contract_income",
            "donation_income",
            "milestone_income",
        ]

        qs = self.filter(character__character__character_id__in=character_ids)

        qs = qs.annotate_ledger_data(exclude)

        annotations = {}
        for type_name in type_names:
            annotations[f"{type_name}_total_amount"] = Coalesce(
                Sum(
                    Case(
                        When(
                            **{
                                f"{type_name}__isnull": False,
                                "date__year": filter_date.year,
                            },
                            then=F(type_name),
                        )
                    )
                ),
                Value(0),
                output_field=DecimalField(),
            )
            annotations[f"{type_name}_total_amount_day"] = Coalesce(
                Sum(
                    Case(
                        When(
                            **{
                                f"{type_name}__isnull": False,
                                "date__year": filter_date.year,
                                "date__month": filter_date.month,
                                "date__day": filter_date.day,
                            },
                            then=F(type_name),
                        )
                    )
                ),
                Value(0),
                output_field=DecimalField(),
            )
            annotations[f"{type_name}_total_amount_hour"] = Coalesce(
                Sum(
                    Case(
                        When(
                            **{
                                f"{type_name}__isnull": False,
                                "date__year": filter_date.year,
                                "date__month": filter_date.month,
                                "date__day": filter_date.day,
                                "date__hour": filter_date.hour,
                            },
                            then=F(type_name),
                        )
                    )
                ),
                Value(0),
                output_field=DecimalField(),
            )

        qs = qs.aggregate(**annotations)

        for type_name in type_names:
            amounts[type_name]["total_amount"] = qs[f"{type_name}_total_amount"]
            amounts[type_name]["total_amount_day"] = qs[f"{type_name}_total_amount_day"]
            amounts[type_name]["total_amount_hour"] = qs[
                f"{type_name}_total_amount_hour"
            ]

        return amounts

    def annotate_billboard(self, chars: list, exclude: list) -> models.QuerySet:
        qs = self.filter(character__character__character_id__in=chars)
        qs = qs.annotate_ledger_data(exclude)
        qs = qs.annotate(
            miscellaneous=Coalesce(
                F("mission_income")
                + F("incursion_income")
                + F("insurance_income")
                + F("market_income")
                + F("contract_income")
                + F("donation_income")
                + F("milestone_income"),
                Value(0),
                output_field=DecimalField(),
            ),
            costs=Coalesce(
                F("market_cost")
                + F("production_cost")
                + F("contract_cost")
                + F("lp_cost")
                + F("traveling_cost")
                + F("asset_cost")
                + F("skill_cost")
                + F("insurance_cost")
                + F("planetary_cost"),
                Value(0),
                output_field=DecimalField(),
            ),
        )
        return qs


class CharWalletManagerBase(models.Manager):
    pass


CharWalletManager = CharWalletManagerBase.from_queryset(CharWalletQuerySet)



"""
Planetary Audit
"""

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.db.models import Q
from django.shortcuts import redirect, render
from django.utils.translation import gettext_lazy as trans
from django.views.decorators.http import require_POST

from ledger.api.helpers import get_alts_queryset, get_character
from ledger.hooks import get_extension_logger
from ledger.models.planetary import CharacterPlanetDetails
from ledger.view_helpers.core import add_info_to_context

logger = get_extension_logger(__name__)


@login_required
@permission_required(["ledger.basic_access"])
def planetary_ledger(request, character_pk):
    context = {
        "character_pk": character_pk,
    }
    context = add_info_to_context(request, context)
    return render(request, "ledger/planetary/planetary_ledger.html", context=context)


@login_required
@permission_required("ledger.basic_access")
def planetary_admin(request):
    """
    Planetary Admin
    """

    context = {}
    context = add_info_to_context(request, context)

    return render(
        request, "ledger/planetary/admin/character_admin.html", context=context
    )


@login_required
@permission_required("ledger.basic_access")
@require_POST
def switch_alarm(request, character_id: list, planet_id: int):
    # Retrieve character_pk from GET parameters
    character_pk = int(request.POST.get("character_pk", 0))

    # Check Permission
    perm, main = get_character(request, character_id)

    if not perm:
        msg = trans("Permission Denied")
        messages.error(request, msg)
        return redirect("ledger:planetary_ledger", character_pk=character_pk)

    if character_id == 0:
        characters = get_alts_queryset(main)
        characters = characters.values_list("character_id", flat=True)
    else:
        characters = [main.character_id]

    filters = Q(planet__character__character__character_id__in=characters)
    if not planet_id == 0:
        filters &= Q(planet__planet__id=planet_id)

    try:
        planets = CharacterPlanetDetails.objects.filter(filters)
        if planets:
            for p in planets:
                p.notification = not p.notification
                p.save()
        else:
            raise CharacterPlanetDetails.DoesNotExist
    except CharacterPlanetDetails.DoesNotExist:
        print("LUL")
        msg = trans("Planet/s not found")
        messages.error(request, msg)
        return redirect("ledger:planetary_ledger", character_pk=character_pk)

    msg = trans("Alarm/s successfully switched")
    messages.info(request, msg)

    return redirect("ledger:planetary_ledger", character_pk=character_pk)



from datetime import datetime
from decimal import Decimal

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q

from allianceauth.eveonline.models import EveCharacter

from ledger import app_settings, models
from ledger.hooks import get_extension_logger

logger = get_extension_logger(__name__)


# TODO Handle it from generate_ledger
def convert_corp_tax(amount: Decimal) -> Decimal:
    """Convert corp tax to correct amount for character ledger"""
    return (amount / app_settings.LEDGER_CORP_TAX) * (
        100 - app_settings.LEDGER_CORP_TAX
    )


def get_character(
    request, character_id, corp=False
) -> tuple[bool, EveCharacter | None]:
    """Get Character and check permissions"""
    perms = True
    if character_id == 0:
        character_id = request.user.profile.main_character.character_id

    try:
        # Corporation View
        if corp:
            main_char = EveCharacter.objects.select_related(
                "character_ownership",
                "character_ownership__user__profile",
                "character_ownership__user__profile__main_character",
            ).get(character_id=request.user.profile.main_character.character_id)
        else:
            main_char = EveCharacter.objects.get(character_id=character_id)
    except ObjectDoesNotExist:
        return False, None

    # check access
    visible = models.CharacterAudit.objects.visible_eve_characters(request.user)
    if main_char not in visible:
        perms = False
    return perms, main_char


def get_corporation(
    request, corporation_id
) -> tuple[bool | None, list[models.CorporationAudit] | None]:
    """Get Corporation and check permissions for each corporation"""
    perms = True
    if corporation_id == 0:
        corporations = get_main_and_alts_ids_corporations(request)
    else:
        corporations = [corporation_id]

    main_corp = models.CorporationAudit.objects.filter(
        corporation__corporation_id__in=corporations
    )

    if not main_corp.exists():
        return None, None

    # Check access
    visible = models.CorporationAudit.objects.visible_to(request.user)
    # Check if there is an intersection between main_corp and visible
    common_corps = main_corp.intersection(visible)
    if not common_corps.exists():
        perms = False
    return perms, main_corp.values_list("corporation__corporation_id", flat=True)


def get_alliance(
    request, alliance_id
) -> tuple[bool | None, list[models.CorporationAudit] | None]:
    """Get Alliance and check permissions for each corporation"""
    perms = True
    if alliance_id == 0:
        alliances = get_main_and_alts_ids_alliances(request)
    else:
        alliances = [alliance_id]

    main_ally = models.CorporationAudit.objects.filter(
        corporation__alliance__alliance_id__in=alliances
    )

    if not main_ally.exists():
        return None, None

    # Check access
    visible = models.CorporationAudit.objects.visible_to(request.user)

    # Check if there is an intersection between main_corp and visible
    common_corps = main_ally.intersection(visible)
    if not common_corps.exists():
        perms = False
    return perms, main_ally.values_list("corporation__alliance__alliance_id", flat=True)


def get_alts_queryset(main_char, corporations=None) -> list[EveCharacter]:
    """Get all alts for a main character, optionally filtered by corporations."""
    try:
        linked_corporations = (
            main_char.character_ownership.user.character_ownerships.all()
        )

        if corporations:
            linked_corporations = linked_corporations.filter(
                character__corporation_id__in=corporations
            )

        linked_corporations = linked_corporations.values_list("character_id", flat=True)

        return EveCharacter.objects.filter(id__in=linked_corporations)
    except ObjectDoesNotExist:
        return EveCharacter.objects.filter(pk=main_char.pk)


def get_corp_alts_queryset(
    main_char, corporations=None
) -> list[models.general.EveEntity]:
    """Get all alts for a main character, optionally filtered by corporations."""
    try:
        linked_characters = (
            main_char.character_ownership.user.character_ownerships.all()
        )
        chars_list = set()

        if corporations:
            linked_characters = linked_characters.filter(
                character__corporation_id__in=corporations
            )

        for char in linked_characters:
            char, _ = models.general.EveEntity.objects.get_or_create(
                eve_id=char.character.character_id,
                defaults={
                    "name": char.character.character_name,
                    "category": "character",
                },
            )
            chars_list.add(char.eve_id)
        return list(models.general.EveEntity.objects.filter(eve_id__in=chars_list))
    except ObjectDoesNotExist:
        chars = EveCharacter.objects.filter(pk=main_char.pk).values("character_id")
        return list(models.general.EveEntity.objects.filter(eve_id__in=chars))


def get_journal_entitys(date: datetime, view, corporations=None) -> set:
    """Get all entity ids from Corporation Journal Queryset filtered by date."""
    filter_date = Q(date__year=date.year)
    if view == "month":
        filter_date &= Q(date__month=date.month)
    elif view == "day":
        filter_date &= Q(date__month=date.month)
        filter_date &= Q(date__day=date.day)

    first_party_ids = models.CorporationWalletJournalEntry.objects.filter(
        filter_date,
        division__corporation__corporation__corporation_id__in=corporations,
    ).values_list("first_party_id", flat=True)

    second_party_ids = models.CorporationWalletJournalEntry.objects.filter(
        filter_date, division__corporation__corporation__corporation_id__in=corporations
    ).values_list("second_party_id", flat=True)

    entity_ids = set(first_party_ids) | set(second_party_ids)

    return entity_ids


def get_main_and_alts_ids_corporations(request) -> set:
    """Get all corporation ids for main and alts."""
    linked_characters = request.user.profile.main_character.character_ownership.user.character_ownerships.select_related(
        "character", "user"
    ).all()

    linked_characters = linked_characters.values_list("character_id", flat=True)
    corp_ids = EveCharacter.objects.filter(id__in=linked_characters).values_list(
        "corporation_id", flat=True
    )

    return set(corp_ids)


def get_main_and_alts_ids_alliances(request) -> set:
    """Get all alliance ids for main and alts."""
    linked_characters = request.user.profile.main_character.character_ownership.user.character_ownerships.select_related(
        "character", "user"
    ).all()

    linked_characters = linked_characters.values_list("character_id", flat=True)
    ally_ids = EveCharacter.objects.filter(id__in=linked_characters).values_list(
        "alliance_id", flat=True
    )

    return set(ally_ids)



from unittest.mock import MagicMock, patch

from django.core.exceptions import ObjectDoesNotExist
from django.test import RequestFactory, TestCase
from esi.models import Token

from allianceauth.corputils.models import CorpMember, CorpStats
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from app_utils.testing import add_character_to_user, create_user_from_evecharacter

from ledger.api.helpers import get_alts_queryset, get_character
from ledger.tests.testdata.load_allianceauth import load_allianceauth

MODULE_PATH = "ledger.api.helpers"


class TestApiHelpers(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        cls.factory = RequestFactory()

        cls.user, cls.char_owner = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
            ],
        )
        cls.user2, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
            ],
        )
        cls.user3, _ = create_user_from_evecharacter(
            1003,
        )
        cls.user4, _ = create_user_from_evecharacter(
            1004,
        )
        cls.user5, _ = create_user_from_evecharacter(
            1005,
        )
        cls.corp = EveCorporationInfo.objects.get(corporation_id=2001)

    def test_get_main_character(self):
        # given
        request = self.factory.get("/")
        request.user = self.user
        # when
        add_character_to_user(
            self.user, EveCharacter.objects.get(character_id=1004), is_main=False
        )
        data = get_character(request, 1004)
        character = EveCharacter.objects.get(character_id=1004)
        # then
        self.assertEqual(data, (True, character))

    def test_get_main_character_no_permission(self):
        # given
        request = self.factory.get("/")
        request.user = self.user3
        # when
        data = get_character(request, 1001)
        character = EveCharacter.objects.get(character_id=1001)
        # then
        self.assertEqual(data, (False, character))

    @patch(MODULE_PATH + ".models.CharacterAudit.objects.visible_eve_characters")
    def test_main_char_not_in_account_chars(self, mock_visible):
        # given
        request = self.factory.get("/")
        mock_visible.return_value = []
        request.user = self.user
        # when
        data = get_character(request, 1001)
        character = EveCharacter.objects.get(character_id=1001)
        # then
        self.assertEqual(data, (False, character))

    def test_get_alts_queryset(self):
        # given
        main_char = EveCharacter.objects.get(character_id=1001)
        # when
        data = get_alts_queryset(main_char)
        # then
        self.assertEqual(data.count(), 1)

    def test_get_alts_queryset_no_linked_characters(self):
        # given
        main_char = MagicMock()
        main_char.character_ownership.user.character_ownerships.all.return_value.values_list.side_effect = (
            ObjectDoesNotExist
        )

        existing_char = EveCharacter.objects.all().first()
        main_char.pk = existing_char.pk

        # when
        data = get_alts_queryset(main_char)
        # then
        self.assertEqual(data.count(), 1)



from django.db.models import Q
from django.test import TestCase
from django.utils import timezone

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import create_user_from_evecharacter

from ledger.models.characteraudit import CharacterWalletJournalEntry
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all

MODULE_PATH = "ledger.managers.characterjournal_manager"


class CharManagerQuerySetTest(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()
        cls.manager = CharacterWalletJournalEntry.objects.all()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
            ],
        )

        cls.char_1 = EveCharacter.objects.get(character_id=1001)
        cls.char_2 = EveCharacter.objects.get(character_id=1002)

        cls.alt_1 = EveCharacter.objects.get(character_id=1003)
        cls.alt_2 = EveCharacter.objects.get(character_id=1004)

    def test_annotate_bounty(self):
        qs = self.manager.annotate_bounty_income()
        self.assertIsNotNone(qs)
        self.assertIn("bounty_income", qs.query.annotations)

    def test_annotate_mission_income(self):
        qs = self.manager.annotate_mission_income()
        self.assertIsNotNone(qs)
        self.assertIn("mission_income", qs.query.annotations)

    def test_annotate_incursion_income(self):
        qs = self.manager.annotate_incursion_income()
        self.assertIsNotNone(qs)
        self.assertIn("incursion_income", qs.query.annotations)

    def test_annotate_market_income(self):
        qs = self.manager.annotate_market_income()
        self.assertIsNotNone(qs)
        self.assertIn("market_income", qs.query.annotations)

    def test_annotate_contract_income(self):
        qs = self.manager.annotate_contract_income()
        self.assertIsNotNone(qs)
        self.assertIn("contract_income", qs.query.annotations)

    def test_annotate_donation_income(self):
        exclude = [4, 5]
        qs = self.manager.annotate_donation_income(exclude)
        self.assertIsNotNone(qs)
        self.assertIn("donation_income", qs.query.annotations)

        qs_no_filter = self.manager.annotate_donation_income()
        self.assertIsNotNone(qs_no_filter)
        self.assertIn("donation_income", qs_no_filter.query.annotations)

    def test_annotate_insurance_income(self):
        qs = self.manager.annotate_insurance_income()
        self.assertIsNotNone(qs)
        self.assertIn("insurance_income", qs.query.annotations)

    def test_annotate_milestone_income(self):
        qs = self.manager.annotate_milestone_income()
        self.assertIsNotNone(qs)
        self.assertIn("milestone_income", qs.query.annotations)

    def test_annotate_contract_cost(self):
        qs = self.manager.annotate_contract_cost()
        self.assertIsNotNone(qs)
        self.assertIn("contract_cost", qs.query.annotations)

    def test_annotate_market_cost(self):
        qs = self.manager.annotate_market_cost()
        self.assertIsNotNone(qs)
        self.assertIn("market_cost", qs.query.annotations)

    def test_annotate_assets_cost(self):
        qs = self.manager.annotate_asset_cost()
        self.assertIsNotNone(qs)
        self.assertIn("asset_cost", qs.query.annotations)

    def test_annotate_traveling_cost(self):
        qs = self.manager.annotate_traveling_cost()
        self.assertIsNotNone(qs)
        self.assertIn("traveling_cost", qs.query.annotations)

    def test_annotate_production_cost(self):
        qs = self.manager.annotate_production_cost()
        self.assertIsNotNone(qs)
        self.assertIn("production_cost", qs.query.annotations)

    def test_annotate_skill_cost(self):
        qs = self.manager.annotate_skill_cost()
        self.assertIsNotNone(qs)
        self.assertIn("skill_cost", qs.query.annotations)

    def test_annotate_insurance_cost(self):
        qs = self.manager.annotate_insurance_cost()
        self.assertIsNotNone(qs)
        self.assertIn("insurance_cost", qs.query.annotations)

    def test_annotate_planetary_cost(self):
        qs = self.manager.annotate_planetary_cost()
        self.assertIsNotNone(qs)
        self.assertIn("planetary_cost", qs.query.annotations)

    def test_annotate_lp_cost(self):
        qs = self.manager.annotate_lp_cost()
        self.assertIsNotNone(qs)
        self.assertIn("lp_cost", qs.query.annotations)

    def test_generate_ledger(self):

        character_ids = [self.char_1, self.char_2]
        filter_date = Q(date__gte=timezone.make_aware(timezone.datetime(2023, 1, 1)))
        exclude = [4, 5]

        # Test with filter
        result_with_filter = self.manager.generate_ledger(
            character_ids, filter_date, exclude
        )
        self.assertIsNotNone(result_with_filter)

    def test_generate_billboard(self):

        character_ids = [1, 2, 3]
        alts = [4, 5]

        # Test with filter
        qs = self.manager.annotate_billboard(character_ids, alts)
        self.assertIsNotNone(qs)
        self.assertIn("bounty_income", qs.query.annotations)
        self.assertIn("miscellaneous", qs.query.annotations)
        self.assertIn("costs", qs.query.annotations)
        self.assertIn("market_cost", qs.query.annotations)
        self.assertIn("production_cost", qs.query.annotations)

    def test_generate_ledger_with_attribute_error(self):
        character_ids = [self.char_1, self.char_2, "a"]
        filter_date = Q(date__gte=timezone.make_aware(timezone.datetime(2023, 1, 1)))
        exclude = [4, 5]

        result_with_filter = self.manager.generate_ledger(
            character_ids, filter_date, exclude
        )
        self.assertIsNotNone(result_with_filter)



from datetime import datetime
from unittest.mock import MagicMock, patch

from ninja import NinjaAPI

from django.test import TestCase
from django.utils import timezone
from eveuniverse.models import EveMarketPrice

from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)
from app_utils.testing import add_character_to_user, create_user_from_evecharacter

from ledger.api.api_helper.character_helper import CharacterProcess
from ledger.api.ledger import LedgerApiEndpoints
from ledger.models.characteraudit import CharacterMiningLedger
from ledger.models.events import Events
from ledger.tests.test_api import _ledgercorpdata
from ledger.tests.test_api._ledgerchardata import (
    CharmonthlyMarch,
    CharmonthlyMarchMulti,
    CharmonthlyMarchWithTaxEvent,
    CharNoMining,
    Charyearly,
    noData,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all

MODULE_PATH = "ledger.api.api_helper.character_helper"


class ManageApiLedgerCharEndpointsTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.advanced_access",
                "ledger.char_audit_admin_manager",
                "ledger.corp_audit_admin_manager",
            ],
        )
        cls.user2, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
                "ledger.advanced_access",
            ],
        )
        cls.user_with_no_permission, _ = create_user_from_evecharacter(
            1003,
            permissions=[
                "ledger.basic_access",
            ],
        )

        cls.user_with_no_data, _ = create_user_from_evecharacter(
            1022,
            permissions=[
                "ledger.basic_access",
                "ledger.advanced_access",
            ],
        )

        cls.api = NinjaAPI()
        cls.manage_api_endpoints = LedgerApiEndpoints(api=cls.api)

    @patch(MODULE_PATH + ".get_alts_queryset")
    def test_get_alts_with_exception(self, mock_get_alts_queryset):
        process = CharacterProcess(
            chars=[], date=timezone.datetime(2024, 3, 19, 0, 0, 0, tzinfo=timezone.utc)
        )
        main = [1]
        mock_get_alts_queryset.side_effect = Exception("Test exception")
        result = process.get_alts(main)
        self.assertIsNone(result)
        mock_get_alts_queryset.assert_called_once_with(1)

    def test_get_ledger_api_with_tax_event(self):
        Events.objects.create(
            title="Test Event",
            date_start=datetime(2024, 3, 19, 0, 0, 0, tzinfo=timezone.utc),
            date_end=datetime(2024, 3, 19, 23, 59, 59, tzinfo=timezone.utc),
            description="This is a test event.",
            char_ledger=True,
            location="Test Location",
        )

        self.client.force_login(self.user)
        url = "/ledger/api/character/0/ledger/date/2024-03-01/view/month/"

        response = self.client.get(url)

        expected_data = CharmonthlyMarchWithTaxEvent
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

        # Corporation
        url = "/ledger/api/corporation/0/ledger/date/2024-03-01/view/month/"

        response = self.client.get(url)
        expected_data = _ledgercorpdata.CorpdatamanyWithTaxEvent

        Events.objects.all().delete()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_ledger_api(self):
        self.client.force_login(self.user)

        fixed_date = timezone.datetime(2025, 1, 31)

        with patch("django.utils.timezone.datetime") as mock_datetime:
            mock_datetime.now.return_value = fixed_date
            mock_datetime.side_effect = lambda *args, **kw: timezone.datetime(
                *args, **kw
            )

            url = "/ledger/api/character/0/ledger/date/2024-03-01/view/month/"

            response = self.client.get(url)

            expected_data = CharmonthlyMarch
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.json(), expected_data)

            # Corporation
            url = "/ledger/api/corporation/0/ledger/date/2024-03-01/view/month/"

            response = self.client.get(url)

            expected_data = _ledgercorpdata.Corpdatamany
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.json(), expected_data)

            # Alliance
            url = "/ledger/api/alliance/0/ledger/date/2024-03-01/view/month/"

            response = self.client.get(url)
            expected_data = _ledgercorpdata.Corpdatamany
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.json(), expected_data)

            # Wrong Type
            url = "/ledger/api/test/0/ledger/date/2024-03-01/view/month/"

            response = self.client.get(url)
            self.assertContains(response, "Wrong Entity Type", status_code=403)

    def test_get_ledger_api_multi(self):
        chars = EveCharacter.objects.filter(character_id__in=[1002, 1003])
        for char in chars:
            add_character_to_user(self.user, char)
        self.client.force_login(self.user)
        url = "/ledger/api/character/0/ledger/date/2024-03-01/view/month/"

        response = self.client.get(url)

        expected_data = CharmonthlyMarchMulti
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_ledger_api_single(self):
        self.client.force_login(self.user)
        url = "/ledger/api/character/1001/ledger/date/2024-03-01/view/month/"
        # when
        response = self.client.get(url)
        # then
        expected_data = CharmonthlyMarch
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

        # Corporation
        url = "/ledger/api/corporation/2001/ledger/date/2024-03-01/view/month/"

        response = self.client.get(url)
        expected_data = _ledgercorpdata.Corpdatamany
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

        # Alliance
        url = "/ledger/api/alliance/3001/ledger/date/2024-03-01/view/month/"

        response = self.client.get(url)

        expected_data = _ledgercorpdata.Corpdatamany
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_ledger_api_year(self):
        # given
        self.client.force_login(self.user)
        url = "/ledger/api/character/1001/ledger/date/2024-01-01/view/year/"
        expected_data = Charyearly
        # when
        response = self.client.get(url)
        # then
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

        # Corporation
        url = "/ledger/api/corporation/2001/ledger/date/2024-01-01/view/year/"

        response = self.client.get(url)

        expected_data = _ledgercorpdata.Corpdatamany
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

        # Alliance
        url = "/ledger/api/alliance/3001/ledger/date/2024-01-01/view/year/"

        response = self.client.get(url)

        expected_data = _ledgercorpdata.Corpdatamany
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_ledger_api_no_permission(self):
        self.client.force_login(self.user_with_no_permission)
        url = "/ledger/api/character/1001/ledger/date/2024-03-01/view/month/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)

        url = "/ledger/api/corporation/2001/ledger/date/2024-03-01/view/month/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)

        url = "/ledger/api/alliance/3001/ledger/date/2024-03-01/view/month/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)

    def test_get_ledger_api_no_data(self):
        # given
        self.client.force_login(self.user_with_no_data)
        url = "/ledger/api/character/0/ledger/date/2024-03-01/view/month/"
        # when
        response = self.client.get(url)
        # then
        expected_data = noData
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

        self.client.force_login(self.user)
        # Corporation
        url = "/ledger/api/corporation/0/ledger/date/2024-12-01/view/month/"
        # when
        response = self.client.get(url)
        # then
        expected_data = _ledgercorpdata.noData
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

        # Alliance
        url = "/ledger/api/alliance/0/ledger/date/2024-12-01/view/month/"
        # when
        response = self.client.get(url)
        # then
        expected_data = _ledgercorpdata.noData
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_ledger_api_amount_is_zero(self):
        # given
        self.client.force_login(self.user_with_no_data)
        url = "/ledger/api/character/0/ledger/date/2024-03-01/view/month/"
        # when
        response = self.client.get(url)
        # then
        expected_data = noData
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_ledger_api_no_mining(self):
        # given
        self.client.force_login(self.user)
        url = "/ledger/api/character/0/ledger/date/2024-03-01/view/month/"
        CharacterMiningLedger.objects.all().delete()
        EveMarketPrice.objects.all().delete()
        # when
        response = self.client.get(url)
        # then
        expected_data = CharNoMining
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    @patch("ledger.models.CorporationWalletJournalEntry.objects.filter")
    def test_get_corporation_ledger_api_single_with_zero_summary_amount(
        self, mock_filter
    ):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/2002/ledger/date/2024-03-01/view/month/"

        # Mock the queryset to return a journal entry with zero summary amount
        mock_entry = MagicMock()
        mock_entry.get.return_value = 0
        mock_entry.get.side_effect = lambda key, default: (
            0 if key in ["total_bounty", "total_ess"] else default
        )
        mock_filter.return_value.select_related.return_value.annotate_ledger.return_value = [
            mock_entry
        ]

        response = self.client.get(url)

        expected_data = _ledgercorpdata.noData
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    @patch("ledger.api.ledger.admin.CharacterAudit.objects.visible_eve_characters")
    def test_get_admin_no_visible(self, mock_visible_to):
        self.client.force_login(self.user_with_no_permission)
        url = "/ledger/api/character/ledger/admin/"

        mock_visible_to.return_value = None

        # when
        response = self.client.get(url)
        # then
        self.assertContains(response, "Permission Denied", status_code=403)

    @patch("ledger.api.ledger.admin.CorporationAudit.objects.visible_to")
    def test_get_corporation_admin_no_visible(self, mock_visible_to):
        self.client.force_login(self.user2)
        # Corporation
        url = "/ledger/api/corporation/ledger/admin/"

        mock_visible_to.return_value = None

        response = self.client.get(url)
        # then
        self.assertContains(response, "Permission Denied", status_code=403)

        # Alliance
        url = "/ledger/api/alliance/ledger/admin/"
        # when
        response = self.client.get(url)
        # then
        self.assertContains(response, "Permission Denied", status_code=403)

    def test_get_admin(self):
        self.client.force_login(self.user2)
        url = "/ledger/api/character/ledger/admin/"

        # when
        response = self.client.get(url)
        # then
        excepted_data = [
            {
                "character": {
                    "1002": {
                        "character_id": 1002,
                        "character_name": "rotze Rotineque",
                        "corporation_id": 2002,
                        "corporation_name": "Eulenclub",
                    }
                }
            }
        ]
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), excepted_data)

        # Corporation
        url = "/ledger/api/corporation/ledger/admin/"
        # when
        response = self.client.get(url)
        # then
        excepted_data = [
            {
                "corporation": {
                    "2002": {"corporation_id": 2002, "corporation_name": "Eulenclub"}
                }
            }
        ]
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), excepted_data)

        # Alliance
        url = "/ledger/api/alliance/ledger/admin/"
        # when
        response = self.client.get(url)
        # then
        excepted_data = [
            {
                "alliance": {
                    "3002": {"alliance_id": 3002, "alliance_name": "Eulen of War"}
                }
            }
        ]
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), excepted_data)

    @patch("ledger.api.ledger.admin.UserProfile.objects.filter")
    def test_get_character_admin_attribute_error(self, mock_user_profile_filter):
        # given
        self.client.force_login(self.user)
        url = "/ledger/api/character/ledger/admin/"

        # Mock the UserProfile to return a character with missing attributes
        mock_user_profile_filter.return_value = [MagicMock(main_character="LUL")]

        # when
        response = self.client.get(url)

        # then
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), [{"character": {}}])

    @patch("ledger.api.ledger.admin.CorporationAudit.objects.visible_to")
    def test_get_admin_exception(self, mock_visible_to):
        self.client.force_login(self.user)
        # Corporation
        url = "/ledger/api/corporation/ledger/admin/"

        corp = EveCorporationInfo.objects.get(corporation_id=2001)

        mock_visible_to.return_value = [corp, "test"]

        # when
        response = self.client.get(url)
        # then
        self.assertEqual(response.status_code, 200)

        # Alliance
        url = "/ledger/api/alliance/ledger/admin/"

        corp = EveAllianceInfo.objects.get(alliance_id=3001)

        mock_visible_to.return_value = [corp, "test"]

        # when
        response = self.client.get(url)
        # then
        self.assertEqual(response.status_code, 200)



from ninja import NinjaAPI

from allianceauth.authentication.models import UserProfile

from ledger.api import schema
from ledger.hooks import get_extension_logger
from ledger.models import CorporationAudit
from ledger.models.characteraudit import CharacterAudit

logger = get_extension_logger(__name__)


class LedgerAdminApiEndpoints:
    tags = ["LedgerAdmin"]

    def __init__(self, api: NinjaAPI):
        @api.get(
            "character/ledger/admin/",
            response={200: list[schema.CharacterAdmin], 403: str},
            tags=self.tags,
        )
        def get_character_admin(request):
            chars_visible = CharacterAudit.objects.visible_eve_characters(request.user)

            if chars_visible is None:
                return 403, "Permission Denied"

            chars_ids = chars_visible.values_list("character_id", flat=True)

            users_char_ids = UserProfile.objects.filter(
                main_character__isnull=False, main_character__character_id__in=chars_ids
            )

            character_dict = {}

            for character in users_char_ids:
                # pylint: disable=broad-exception-caught
                try:
                    character_dict[character.main_character.character_id] = {
                        "character_id": character.main_character.character_id,
                        "character_name": character.main_character.character_name,
                        "corporation_id": character.main_character.corporation_id,
                        "corporation_name": character.main_character.corporation_name,
                    }
                except AttributeError:
                    continue

            output = []
            output.append({"character": character_dict})

            return output

        @api.get(
            "character/planetary/admin/",
            response={200: list[schema.CharacterAdmin], 403: str},
            tags=self.tags,
        )
        def get_planetary_admin(request):
            chars_ids = CharacterAudit.objects.visible_eve_characters(
                request.user
            ).values_list("character_id", flat=True)

            users_char_ids = UserProfile.objects.filter(
                main_character__isnull=False, main_character__character_id__in=chars_ids
            )

            if not chars_ids:
                return 403, "Permission Denied"

            character_dict = {}

            for character in users_char_ids:
                # pylint: disable=broad-exception-caught
                try:
                    character_dict[character.main_character.character_id] = {
                        "character_id": character.main_character.character_id,
                        "character_name": character.main_character.character_name,
                        "corporation_id": character.main_character.corporation_id,
                        "corporation_name": character.main_character.corporation_name,
                    }
                except Exception:
                    continue

            output = []
            output.append({"character": character_dict})

            return output

        @api.get(
            "corporation/ledger/admin/",
            response={200: list[schema.CorporationAdmin], 403: str},
            tags=self.tags,
        )
        def get_corporation_admin(request):
            corporations = CorporationAudit.objects.visible_to(request.user)

            if corporations is None:
                return 403, "Permission Denied"

            corporation_dict = {}

            for corporation in corporations:
                # pylint: disable=broad-exception-caught
                try:
                    corporation_dict[corporation.corporation.corporation_id] = {
                        "corporation_id": corporation.corporation.corporation_id,
                        "corporation_name": corporation.corporation.corporation_name,
                    }
                except Exception:
                    continue

            output = []
            output.append({"corporation": corporation_dict})

            return output

        @api.get(
            "alliance/ledger/admin/",
            response={200: list[schema.AllianceAdmin], 403: str},
            tags=self.tags,
        )
        def get_alliance_admin(request):
            corporations = CorporationAudit.objects.visible_to(request.user)

            if corporations is None:
                return 403, "Permission Denied"

            alliance_dict = {}

            for corporation in corporations:
                # pylint: disable=broad-exception-caught
                try:
                    logger.debug(corporation)
                    alliance_dict[corporation.corporation.alliance.alliance_id] = {
                        "alliance_id": corporation.corporation.alliance.alliance_id,
                        "alliance_name": corporation.corporation.alliance.alliance_name,
                    }
                    logger.debug(alliance_dict)
                except Exception:
                    continue
            logger.debug(alliance_dict)
            output = []
            output.append({"alliance": alliance_dict})

            return output



from datetime import datetime

from ninja import Schema

from ledger.api.api_helper.billboard_helper import BillboardDict


class Message(Schema):
    message: str


class Character(Schema):
    character_name: str
    character_id: int
    corporation_id: int
    corporation_name: str
    alliance_id: int | None = None
    alliance_name: str | None = None


class Corporation(Schema):
    corporation_id: int
    corporation_name: str
    alliance_id: int | None = None
    alliance_name: str | None = None


class EveName(Schema):
    id: int
    name: str
    cat: str | None = None


class CharacterWalletEvent(Schema):
    character: Character | None = None
    id: int | None = None
    date: datetime | None = None
    first_party: EveName | None = None
    second_party: EveName | None = None
    ref_type: str | None = None
    balance: float | None = None
    amount: float | None = None
    reason: str | None = None


class CharacterPlanet(Schema):
    character_id: int | None = None
    character_name: str | None = None
    planet: str | None = None
    planet_id: int | None = None
    upgrade_level: int | None = None
    num_pins: int | None = None
    last_update: datetime | None = None


class CharacterPlanetDetails(Schema):
    character_id: int | None = None
    character_name: str | None = None
    planet: str | None = None
    planet_id: int | None = None
    planet_type_id: int | None = None
    upgrade_level: int | None = None
    expiry_date: datetime | None = None
    expired: bool | None = None
    alarm: bool | None = None
    products: dict | None = None
    products_info: dict | None = None
    extractors: dict | None = None
    last_update: datetime | None = None
    facility: dict | None = None


class CharacterLedger(Schema):
    ratting: list | None = None
    total: dict | None = None
    billboard: BillboardDict | None = None


class CharacterLedgerTemplate(Schema):
    character: str | None = None


class CharacterAdmin(Schema):
    character: dict | None = None


class CorporationAdmin(Schema):
    corporation: dict | None = None


class AllianceAdmin(Schema):
    alliance: dict | None = None


class CorporationWalletEvent(Schema):
    division: str
    id: int
    date: datetime
    first_party: EveName
    second_party: EveName
    ref_type: str
    balance: float
    amount: float
    reason: str | None = None


class Ledger(Schema):
    ratting: list | None = None
    total: dict | None = None


class Billboard(Schema):
    billboard: BillboardDict | None = None



from unittest.mock import Mock, patch

from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase
from django.urls import reverse
from esi.models import Token

from allianceauth.eveonline.providers import ObjectNotFound
from app_utils.testing import create_user_from_evecharacter

from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.views.alliance.ally_audit import add_ally

MODULE_PATH = "ledger.views.alliance.ally_audit"


class CharAuditTest(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.admin_access",
                "ledger.corp_audit_admin_manager",
            ],
        )

        cls.token = Mock(spec=Token)
        cls.token.character_id = cls.character_ownership.character.character_id
        cls.token.corporation_id = cls.character_ownership.character.corporation_id
        cls.token.corporation_name = cls.character_ownership.character.corporation_name
        cls.token.corporation_ticker = (
            cls.character_ownership.character.corporation_ticker
        )
        cls.token.alliance_id = 3001
        cls.token.alliance_name = "Voices of War"
        cls.token.alliance_ticker = "VOW"

    @patch(MODULE_PATH + ".messages")
    def test_add_ally(self, mock_messages):
        self.client.force_login(self.user)
        request = self.factory.get(reverse("ledger:ledger_add_ally"))
        request.user = self.user
        request.token = self.token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # given
        orig_view = add_ally.__wrapped__.__wrapped__
        # when
        response = orig_view(request, self.token)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response.url, reverse("ledger:alliance_ledger", kwargs={"alliance_pk": 0})
        )
        self.assertTrue(mock_messages.info.called)

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".EveAllianceInfo.objects.filter")
    @patch(MODULE_PATH + ".provider.get_alliance")
    @patch(MODULE_PATH + ".EveAllianceInfo.objects.get_or_create")
    def test_add_ally_not_exist_create_ally(
        self, mock_get_or_create, mock_get_alliance, mock_ally, mock_messages
    ):
        self.client.force_login(self.user)

        mock_get_alliance.return_value = Mock()
        mock_get_alliance.return_value.id = 9999
        mock_get_alliance.return_value.name = "Voices of War"
        mock_get_alliance.return_value.ticker = "VOW"
        mock_get_alliance.return_value.executor_corp_id = 2001

        mock_ally.return_value.first.return_value = None

        mock_get_or_create.return_value = (Mock(), True)
        mock_ally.populate_alliance.return_value = None

        request = self.factory.get(reverse("ledger:ledger_add_ally"))
        request.user = self.user
        request.token = self.token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # given
        orig_view = add_ally.__wrapped__.__wrapped__
        # when
        response = orig_view(request, self.token)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response.url, reverse("ledger:alliance_ledger", kwargs={"alliance_pk": 0})
        )
        self.assertTrue(mock_messages.success.called)

    @patch(MODULE_PATH + ".messages")
    @patch(MODULE_PATH + ".EveAllianceInfo.objects.filter")
    @patch(MODULE_PATH + ".provider.get_alliance")
    def test_add_ally_not_exist_object_not_found(
        self, mock_get_alliance, mock_ally, mock_messages
    ):
        self.client.force_login(self.user)

        mock_get_alliance.side_effect = ObjectNotFound(3001, "Alliance not found")

        mock_ally.return_value.first.return_value = None

        request = self.factory.get(reverse("ledger:ledger_add_ally"))
        request.user = self.user
        request.token = self.token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        # given
        orig_view = add_ally.__wrapped__.__wrapped__
        # when
        response = orig_view(request, self.token)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response.url, reverse("ledger:alliance_ledger", kwargs={"alliance_pk": 0})
        )
        self.assertTrue(mock_messages.warning.called)



from unittest.mock import MagicMock, patch

from ninja import NinjaAPI

from django.test import TestCase
from django.utils import timezone

from app_utils.testing import create_user_from_evecharacter

from ledger.api.ledger.planetary import LedgerPlanetaryApiEndpoints
from ledger.tests.test_api import _planetchardata
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all
from ledger.tests.testdata.load_planetary import load_planetary


class ManageApiJournalCharEndpointsTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()
        load_planetary()

        cls.user, _ = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
            ],
        )

        cls.user2, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
            ],
        )
        cls.api = NinjaAPI()
        cls.manage_api_endpoints = LedgerPlanetaryApiEndpoints(api=cls.api)

    def test_get_character_planetary_api(self):
        self.client.force_login(self.user)
        url = "/ledger/api/character/0/planetary/0/"

        response = self.client.get(url)
        expected_data = [
            {
                "character_id": 1001,
                "character_name": "Gneuten",
                "planet": "Test Planet I",
                "planet_id": 4001,
                "upgrade_level": 5,
                "num_pins": 5,
                "last_update": None,
            },
            {
                "character_id": 1001,
                "character_name": "Gneuten",
                "planet": "Test Planet I",
                "planet_id": 4002,
                "upgrade_level": 5,
                "num_pins": 5,
                "last_update": None,
            },
        ]

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_planetary_api_single(self):
        self.client.force_login(self.user)
        url = "/ledger/api/character/1001/planetary/0/"

        response = self.client.get(url)

        expected_data = [
            {
                "character_id": 1001,
                "character_name": "Gneuten",
                "planet": "Test Planet I",
                "planet_id": 4001,
                "upgrade_level": 5,
                "num_pins": 5,
                "last_update": None,
            },
            {
                "character_id": 1001,
                "character_name": "Gneuten",
                "planet": "Test Planet I",
                "planet_id": 4002,
                "upgrade_level": 5,
                "num_pins": 5,
                "last_update": None,
            },
        ]

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_planetary_api_single_planet(self):
        self.client.force_login(self.user)
        url = "/ledger/api/character/1001/planetary/4001/"

        response = self.client.get(url)

        expected_data = [
            {
                "character_id": 1001,
                "character_name": "Gneuten",
                "planet": "Test Planet I",
                "planet_id": 4001,
                "upgrade_level": 5,
                "num_pins": 5,
                "last_update": None,
            },
        ]

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_planetary_api_no_permission(self):
        self.client.force_login(self.user2)
        url = "/ledger/api/character/1001/planetary/0/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), "Permission Denied")

    def test_get_character_planetary_details_api(self):
        self.client.force_login(self.user)
        url = "/ledger/api/character/0/planetary/0/details/"

        response = self.client.get(url)
        expected_data = _planetchardata.planet_many

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_planetary_details_all_planets(self):
        self.client.force_login(self.user)
        url = "/ledger/api/character/1001/planetary/0/details/"

        response = self.client.get(url)

        expected_data = _planetchardata.planet_many

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_planetary_details_single_planet_with_no_expired_mock(self):
        self.client.force_login(self.user)
        url = "/ledger/api/character/1001/planetary/4001/details/"

        # Set the fixed date as a timezone-aware datetime object
        fixed_date = timezone.make_aware(timezone.datetime(2024, 8, 20, 17, 17, 2))

        # Mock the timezone.now() method to return the fixed date
        with patch("django.utils.timezone.now", return_value=fixed_date):
            response = self.client.get(url)
            expected_data = _planetchardata.planet_single

            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.json(), expected_data)

    def test_get_character_planetary_details_api_no_permission(self):
        self.client.force_login(self.user2)
        url = "/ledger/api/character/1001/planetary/0/details/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), "Permission Denied")

    @patch("ledger.api.ledger.admin.CharacterAudit.objects.visible_eve_characters")
    def test_get_character_admin_planetary_no_visible(self, mock_visible_to):
        self.client.force_login(self.user2)
        url = "/ledger/api/character/planetary/admin/"

        mock_visible_to.return_value.values_list.return_value = []

        # when
        response = self.client.get(url)
        # then
        self.assertContains(response, "Permission Denied", status_code=403)

    def test_get_character_planetary_admin(self):
        self.client.force_login(self.user2)
        url = "/ledger/api/character/planetary/admin/"

        # when
        response = self.client.get(url)
        # then
        excepted_data = [
            {
                "character": {
                    "1002": {
                        "character_id": 1002,
                        "character_name": "rotze Rotineque",
                        "corporation_id": 2002,
                        "corporation_name": "Eulenclub",
                    }
                }
            }
        ]
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), excepted_data)

    @patch("ledger.api.ledger.admin.UserProfile.objects.filter")
    def test_get_character_admin_planetary_attribute_error(
        self, mock_user_profile_filter
    ):
        # given
        self.client.force_login(self.user)
        url = "/ledger/api/character/planetary/admin/"

        # Mock the UserProfile to return a character with missing attributes
        mock_user_profile_filter.return_value = [MagicMock(main_character="LUL")]

        # when
        response = self.client.get(url)

        # then
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), [{"character": {}}])



from unittest.mock import MagicMock, call, patch

from django.core.cache import cache
from django.test import TestCase
from django.utils import timezone
from esi.models import Token
from eveuniverse.models import EvePlanet, EveSolarSystem, EveType

from ledger.models.characteraudit import CharacterAudit
from ledger.models.planetary import CharacterPlanet, CharacterPlanetDetails
from ledger.task_helpers.etag_helpers import NotModifiedError
from ledger.task_helpers.plan_helpers import (
    convert_datetime_to_str,
    update_character_planetary,
    update_character_planetary_details,
)
from ledger.tasks import update_char_planets_details
from ledger.tests.testdata.esi_stub import esi_client_stub
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all
from ledger.tests.testdata.load_planetary import load_planetary
from ledger.tests.testdata.planetary import planetary_data

MODULE_PATH = "ledger.task_helpers.plan_helpers"


class TestCharacterPlanetaryHelpers(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()
        load_planetary()

        cls.mock_token = MagicMock(spec=Token)
        cls.mock_token.character_id = 1001
        cls.mock_token.valid_access_token.return_value = "token"

        cls.planetary = [
            {
                "planet_id": 4001,
                "upgrade_level": 5,
                "num_pins": 10,
            },
            {
                "planet_id": 4002,
                "upgrade_level": 4,
                "num_pins": 8,
            },
        ]

        cls.planetarydetails = planetary_data

        cache.clear()

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".CharacterPlanet.objects.filter")
    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".EvePlanet.objects.get_or_create_esi")
    @patch(MODULE_PATH + ".CharacterPlanet.objects.bulk_update")
    def test_update_character_planetary(
        self,
        mock_update,
        mock_get_or_create_esi,
        mock_etag,
        mock_esi,
        mock_planet_details,
        mock_planet,
        mock_get_token,
    ):
        # given
        mock_get_token.return_value = self.mock_token
        mock_esi.client = esi_client_stub
        # TODO get Data from Test ESI
        # mock_etag.return_value = mock_esi.client.Wallet.get_characters_character_id_wallet_journal(character_id=1001)
        mock_etag.return_value = self.planetary

        mock_get_or_create_esi.side_effect = [
            (EvePlanet.objects.get(id=4001), False),  # Existing planet
            (
                EvePlanet.objects.create(
                    id=4003,
                    name="Test Planet III",
                    last_updated="2021-01-01T00:00:00Z",
                    position_x=0.0,
                    position_y=0.0,
                    position_z=0.0,
                    eve_solar_system=EveSolarSystem.objects.get(id=30004783),
                    eve_type=EveType.objects.get(id=13),
                    enabled_sections=0,
                ),
                True,
            ),  # New planet
        ]
        mock_planet.return_value.values_list.return_value = [4001]

        # when
        result = update_character_planetary(1001)

        # then
        self.assertEqual(result, ("Finished planets update for: %s", "Gneuten"))
        self.assertEqual(mock_update.call_count, 1)
        self.assertEqual(mock_planet.filter.return_value.delete.call_count, 0)
        self.assertEqual(mock_planet_details.filter.return_value.delete.call_count, 1)

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".CharacterPlanet.objects.filter")
    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".EvePlanet.objects.get_or_create_esi")
    @patch(MODULE_PATH + ".CharacterPlanet.objects.bulk_update")
    def test_update_character_planetary_exist(
        self,
        mock_planet_update,
        mock_get_or_create_esi,
        mock_etag,
        mock_esi,
        mock_planet_details,
        mock_planet,
        mock_get_token,
    ):
        # given
        mock_get_token.return_value = self.mock_token
        mock_esi.client = esi_client_stub
        # TODO get Data from Test ESI
        # mock_etag.return_value = mock_esi.client.Wallet.get_characters_character_id_wallet_journal(character_id=1001)
        mock_etag.return_value = self.planetary
        mock_get_or_create_esi.return_value = (EvePlanet.objects.get(id=4001), False)
        mock_planet.return_value.values_list.return_value = [4001]

        # when
        result = update_character_planetary(1001)

        # then
        self.assertEqual(result, ("Finished planets update for: %s", "Gneuten"))
        self.assertEqual(mock_planet_update.call_count, 1)
        self.assertEqual(mock_planet.filter.return_value.delete.call_count, 0)
        self.assertEqual(mock_planet_details.filter.return_value.delete.call_count, 1)

    @patch(MODULE_PATH + ".get_token")
    def test_update_character_planetary_no_token(self, mock_get_token):
        # given
        mock_get_token.return_value = None
        # when
        result = update_character_planetary(1001)
        # then
        self.assertEqual(result, ("No Tokens"))

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".CharacterPlanet.objects.filter")
    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".EvePlanet.objects.get_or_create_esi")
    @patch(MODULE_PATH + ".CharacterPlanet.objects.bulk_update")
    def test_update_character_planetary_existing_planets(
        self,
        mock_planet_update,
        mock_get_or_create_esi,
        mock_etag,
        mock_esi,
        mock_planet_details,
        mock_planet,
        mock_get_token,
    ):
        # given
        mock_get_token.return_value = self.mock_token
        mock_esi.client = MagicMock()
        mock_etag.return_value = [
            {"planet_id": 4001, "upgrade_level": 1, "num_pins": 5},
            {"planet_id": 4002, "upgrade_level": 2, "num_pins": 10},
        ]
        mock_planet.return_value.values_list.return_value = [4001, 4002]

        mock_get_or_create_esi.side_effect = [
            (EvePlanet(id=4001, name="Planet 1"), False),
            (EvePlanet(id=4002, name="Planet 2"), False),
        ]

        # when
        result = update_character_planetary(1001)

        # then
        self.assertEqual(result, ("Finished planets update for: %s", "Gneuten"))
        self.assertEqual(mock_planet_update.call_count, 1)

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".CharacterPlanet.objects.filter")
    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".EvePlanet.objects.get_or_create_esi")
    @patch(MODULE_PATH + ".CharacterPlanet.objects.get")
    @patch(MODULE_PATH + ".CharacterPlanet.objects.bulk_create")
    def test_update_character_planetary_planet_not_exist(
        self,
        mock_planet_create,
        mock_get_planet,
        mock_get_or_create_esi,
        mock_etag,
        mock_esi,
        mock_planet_details,
        mock_planet,
        mock_get_token,
    ):
        # given
        mock_get_token.return_value = self.mock_token
        mock_esi.client = MagicMock()
        mock_etag.return_value = [
            {"planet_id": 4004, "upgrade_level": 1, "num_pins": 5},
        ]
        mock_planet.return_value.values_list.return_value = []

        mock_get_or_create_esi.side_effect = [
            (
                EvePlanet.objects.create(
                    id=4004,
                    name="Planet 1",
                    last_updated="2021-01-01T00:00:00Z",
                    position_x=0.0,
                    position_y=0.0,
                    position_z=0.0,
                    eve_solar_system=EveSolarSystem.objects.get(id=30004783),
                    eve_type=EveType.objects.get(id=13),
                    enabled_sections=0,
                ),
                True,
            ),
        ]

        mock_get_planet.side_effect = CharacterPlanet.DoesNotExist

        # when
        result = update_character_planetary(1001)

        # then
        self.assertEqual(result, ("Finished planets update for: %s", "Gneuten"))
        self.assertEqual(mock_planet_create.call_count, 1)

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".logger")
    def test_update_character_planetary_not_modified(
        self,
        mock_logger,
        mock_etag,
        mock_esi,
        mock_get_token,
    ):
        # given
        mock_get_token.return_value = self.mock_token
        mock_esi.client = MagicMock()
        mock_etag.side_effect = NotModifiedError

        # when
        result = update_character_planetary(1002)

        # then
        self.assertEqual(result, ("Finished planets update for: %s", "rotze Rotineque"))
        mock_logger.debug.assert_called_with(
            "No New Planet data for: %s", "rotze Rotineque"
        )


class TestCharacterPlanetaryDetailsHelpers(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()
        load_planetary()

        cls.mock_token = MagicMock(spec=Token)
        cls.mock_token.character_id = 1001
        cls.mock_token.valid_access_token.return_value = "token"

        cls.planetary = [
            {
                "planet_id": 4001,
                "upgrade_level": 5,
                "num_pins": 10,
            },
            {
                "planet_id": 4002,
                "upgrade_level": 4,
                "num_pins": 8,
            },
        ]

        cls.planetarydetails = planetary_data

    @patch(MODULE_PATH + ".get_token")
    def test_update_character_planetarydetails_no_token(self, mock_get_token):
        # given
        mock_get_token.return_value = None
        # when
        result = update_character_planetary_details(1001, 4001)
        # then
        self.assertEqual(result, ("No Tokens"))

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects.update_or_create_layout")
    def test_update_character_planetarydetails_exist(
        self,
        mock_update_or_create,
        mock_etag,
        mock_esi,
        mock_get_token,
    ):
        # given
        mock_get_token.return_value = self.mock_token
        mock_esi.client.return_value = "Test ESI"
        # TODO get Data from Test ESI
        mock_etag.return_value = self.planetarydetails

        mock_update_or_create.return_value = (
            CharacterPlanetDetails(
                planet=CharacterPlanet.objects.get(
                    character__character__character_id=1001,
                    planet_id=4001,
                ),
                links=self.planetarydetails["links"],
                pins=self.planetarydetails["pins"],
                routes=self.planetarydetails["routes"],
                last_update=timezone.now(),
            ),
            False,
        )

        # when
        result = update_character_planetary_details(1001, 4001)

        # then
        self.assertEqual(result, ("Finished planets details update for: %s", "Gneuten"))

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects.update_or_create_layout")
    def test_update_character_planetarydetails_new_details(
        self,
        mock_update_or_create,
        mock_etag,
        mock_esi,
        mock_get_token,
    ):
        # given
        mock_get_token.return_value = self.mock_token
        mock_esi.client.return_value = "Test ESI"
        # TODO get Data from Test ESI
        mock_etag.return_value = self.planetarydetails

        planet = EvePlanet.objects.create(
            id=4099,
            name="Planet 1",
            last_updated="2021-01-01T00:00:00Z",
            position_x=0.0,
            position_y=0.0,
            position_z=0.0,
            eve_solar_system=EveSolarSystem.objects.get(id=30004783),
            eve_type=EveType.objects.get(id=13),
            enabled_sections=0,
        )

        CharacterPlanet.objects.create(
            character=CharacterAudit.objects.get(character__character_id=1001),
            planet=planet,
            upgrade_level=5,
            num_pins=5,
        )

        mock_update_or_create.return_value = (
            CharacterPlanetDetails(
                planet=CharacterPlanet.objects.get(
                    character__character__character_id=1001,
                    planet_id=4099,
                ),
                links=self.planetarydetails["links"],
                pins=self.planetarydetails["pins"],
                routes=self.planetarydetails["routes"],
                last_update=timezone.now(),
            ),
            True,
        )

        # when
        result = update_character_planetary_details(1001, 4099)

        # then
        self.assertEqual(result, ("Finished planets details update for: %s", "Gneuten"))

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects.update_or_create_layout")
    @patch(MODULE_PATH + ".logger")
    def test_update_character_planetarydetails_expired_no_alert(
        self,
        mock_logger,
        mock_update_or_create,
        mock_etag,
        mock_esi,
        mock_get_token,
    ):
        # given
        mock_get_token.return_value = self.mock_token
        mock_esi.client.Planetary_Interaction.get_characters_character_id_planets_planet_id.return_value = (
            self.planetarydetails
        )
        mock_etag.return_value = self.planetarydetails

        mock_update_or_create.return_value = (
            CharacterPlanetDetails(
                planet=CharacterPlanet.objects.get(
                    character__character__character_id=1001,
                    planet_id=4001,
                ),
                links=self.planetarydetails["links"],
                pins=self.planetarydetails["pins"],
                routes=self.planetarydetails["routes"],
                last_update=timezone.now(),
                last_alert=None,
            ),
            False,
        )

        # when
        result = update_character_planetary_details(1001, 4001)

        # then
        self.assertEqual(result, ("Finished planets details update for: %s", "Gneuten"))
        mock_logger.debug.assert_called_with(
            "Planet %s Extractor Heads Expired for: %s", "Test Planet I", "Gneuten"
        )

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects.update_or_create_layout")
    @patch(MODULE_PATH + ".logger")
    def test_update_character_planetarydetails_alert_reset(
        self,
        mock_logger,
        mock_update_or_create,
        mock_etag,
        mock_esi,
        mock_get_token,
    ):
        # given
        mock_get_token.return_value = self.mock_token
        mock_esi.client.Planetary_Interaction.get_characters_character_id_planets_planet_id.return_value = (
            self.planetarydetails
        )
        mock_etag.return_value = self.planetarydetails

        mock_update_or_create.return_value = (
            CharacterPlanetDetails(
                planet=CharacterPlanet.objects.get(
                    character__character__character_id=1001,
                    planet_id=4002,
                ),
                links=self.planetarydetails["links"],
                pins=self.planetarydetails["pins"],
                routes=self.planetarydetails["routes"],
                last_update=timezone.now(),
                last_alert=timezone.now() - timezone.timedelta(days=2),
            ),
            False,
        )

        # when
        result = update_character_planetary_details(1001, 4002)

        # then
        self.assertEqual(result, ("Finished planets details update for: %s", "Gneuten"))
        mock_logger.debug.assert_called_with(
            "Notification Reseted for %s Planet: %s", "Gneuten", "Test Planet I"
        )

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".logger")
    def test_update_character_planetarydetails_not_modified(
        self,
        mock_logger,
        mock_etag,
        mock_esi,
        mock_get_token,
    ):
        # given
        mock_get_token.return_value = self.mock_token
        mock_esi.client = MagicMock()
        mock_etag.side_effect = NotModifiedError

        # when
        result = update_character_planetary_details(1001, 4001)

        # then
        self.assertEqual(result, ("No New Planet Details data for: %s", "Gneuten"))
        mock_logger.debug.assert_called_with(
            "No New Planet Details data for: %s", "Gneuten"
        )

    def test_convert_datetime_to_str(self):
        # given
        sample_datetime = timezone.now()
        sample_data = {"name": "test", "timestamp": sample_datetime}

        # when
        result = convert_datetime_to_str(sample_data)

        # then
        self.assertEqual(result["timestamp"], sample_datetime.isoformat())



from .admin import LedgerAdminApiEndpoints
from .ledger import LedgerApiEndpoints
from .planetary import LedgerPlanetaryApiEndpoints
from .template import LedgerTemplateApiEndpoints


def setup(api):
    LedgerAdminApiEndpoints(api)
    LedgerApiEndpoints(api)
    LedgerPlanetaryApiEndpoints(api)
    LedgerTemplateApiEndpoints(api)



from datetime import datetime

from django.db.models import Q

from ledger.api.api_helper.billboard_helper import BillboardLedger
from ledger.api.api_helper.core_manager import LedgerModels, LedgerTotal
from ledger.hooks import get_extension_logger
from ledger.models.corporationaudit import CorporationWalletJournalEntry

logger = get_extension_logger(__name__)


class CorporationProcess:
    """JournalProcess class to process the journal entries."""

    def __init__(self, corporations, date: datetime, view=None):
        self.corp = corporations if corporations else []
        self.date = date
        self.view = view

    def _filter_date(self):
        """Filter the date."""
        filter_date = Q(date__year=self.date.year)
        if self.view == "month":
            filter_date &= Q(date__month=self.date.month)
        elif self.view == "day":
            filter_date &= Q(date__month=self.date.month)
            filter_date &= Q(date__day=self.date.day)
        return filter_date

    # pylint: disable=too-many-locals
    def _process_corporation_chars(self, journal):
        # Create the Dicts for each Character
        corporation_dict = {}
        corporation_total = LedgerTotal()

        # Annotate Data
        journal = (
            journal.annotate_bounty_income()
            .annotate_ess_income()
            .annotate_miscellaneous()
        )
        for main in journal:
            total_bounty = main.get("bounty_income", 0)
            total_ess = main.get("ess_income", 0)
            total_other = main.get("miscellaneous", 0)
            main_entity_id = main.get("main_entity_id", 0)
            character_name = main.get("main_entity_name") or "Unknown"
            alts = main.get("alts", [])
            entity_type = "character"

            summary_amount = sum([total_bounty, total_ess, total_other])

            if summary_amount > 0:
                corporation_dict[main_entity_id] = {
                    "main_id": main_entity_id,
                    "main_name": character_name,
                    "entity_type": entity_type,
                    "alt_names": alts,
                    "total_amount": total_bounty,
                    "total_amount_ess": total_ess,
                    "total_amount_others": total_other,
                }

            totals = {
                "total_amount": total_bounty,
                "total_amount_ess": total_ess,
                "total_amount_others": total_other,
                "total_amount_all": summary_amount,
            }
            # Summary all
            corporation_total.get_summary(totals)

        return corporation_dict, corporation_total

    def generate_ledger(self):
        # Get the Filter Settings
        filter_date = self._filter_date()

        journal = (
            CorporationWalletJournalEntry.objects.filter(filter_date)
            .select_related(
                "first_party",
                "second_party",
            )
            .generate_ledger(self.corp)
        )

        # Create the Dicts for each Character
        corporation_dict, corporation_total = self._process_corporation_chars(journal)

        output = []
        output.append(
            {
                "ratting": sorted(
                    list(corporation_dict.values()), key=lambda x: x["main_name"]
                ),
                "total": corporation_total.to_dict(),
            }
        )

        return output

    def generate_billboard(self, corporations):
        # Get the Filter Settings
        filter_date = self._filter_date()

        corporation_journal = (
            CorporationWalletJournalEntry.objects.filter(filter_date)
            .select_related(
                "first_party",
                "second_party",
            )
            .generate_ledger(corporations)
        )

        chars_list = list(corporation_journal.values_list("second_party_id", flat=True))

        models = LedgerModels(corporation_journal=corporation_journal)

        # Create the Billboard for the Corporation
        ledger = BillboardLedger(view=self.view, models=models, corp=True)

        billboard_dict = ledger.billboard_ledger(chars_list)

        output = []
        output.append(
            {
                "billboard": billboard_dict,
            }
        )

        return output



from django.utils.translation import gettext as trans


def get_facilities_info(planet):
    facilities = []
    for _, facility_info in planet.facilitys.items():
        resources = facility_info["resources"]
        resource_counts = get_resource_counts(resources)

        input_icons = [
            {
                "icon_url": resource["item_id"],
                "item_name": resource["item_name"],
                "count": resource["count"],
            }
            for resource in resource_counts.values()
        ]

        output_icon = (
            {
                "icon_url": facility_info["output_product"]["item_id"],
                "item_name": facility_info["output_product"]["item_name"],
            }
            if facility_info["output_product"]
            else None
        )

        producing = any(resource["still_producing"] for resource in resources)
        is_active = not any(resource["missing_quantity"] > 0 for resource in resources)
        if producing and any(
            resource["missing_quantity"] > 0 for resource in resources
        ):
            is_active = True

        facilities.append(
            {
                "facility_name": facility_info["facility_name"] or trans("No facility"),
                "input_icons": input_icons,
                "output_icon": output_icon,
                "is_active": is_active,
            }
        )
    return facilities


def get_resource_counts(resources):
    resource_counts = {}
    for resource in resources:
        if resource["item_id"] not in resource_counts:
            resource_counts[resource["item_id"]] = {
                "item_id": resource["item_id"],
                "item_name": resource["item_name"],
                "count": 0,
            }
        resource_counts[resource["item_id"]]["count"] += 1
    return resource_counts



from datetime import timedelta
from unittest.mock import MagicMock, call, patch

from django.db import IntegrityError
from django.test import TestCase
from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _
from esi.errors import TokenExpiredError
from eveuniverse.models import EvePlanet

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import add_character_to_user, create_user_from_evecharacter

from ledger.models.characteraudit import CharacterAudit
from ledger.models.corporationaudit import CorporationAudit
from ledger.tasks import (
    check_planetary_alarms,
    update_all_characters,
    update_all_corps,
    update_char_mining_ledger,
    update_char_planets,
    update_char_planets_details,
    update_char_wallet,
    update_character,
    update_corp,
    update_corp_wallet,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all
from ledger.tests.testdata.load_planetary import load_planetary

MODULE_PATH = "ledger.tasks"


class TestTasks(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()
        load_planetary()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
            ],
        )

        cls.user2, cls.character_ownership2 = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
            ],
        )

        cls.token = cls.user.token_set.first()
        cls.corporation = cls.character_ownership.character.corporation

    @patch(MODULE_PATH + ".update_character.apply_async")
    @patch(MODULE_PATH + ".logger")
    def test_update_all_characters(self, mock_logger, mock_update_character):
        # given
        characters_count = CharacterAudit.objects.count()
        # when
        update_all_characters()
        # then
        self.assertEqual(mock_update_character.call_count, characters_count)
        mock_logger.info.assert_called_once_with(
            "Queued %s Char Audit Updates", characters_count
        )

    @patch(MODULE_PATH + ".CharacterAudit.objects.select_related")
    def test_update_character(self, mock_select_related):
        # given
        mock_character = MagicMock()
        mock_character.last_update_mining = timezone.now()
        mock_character.last_update_wallet = timezone.now()
        mock_character.last_update_planetary = timezone.now()
        mock_character.character = MagicMock()
        mock_character.character.character_id = self.token.character_id
        mock_character.character.character_name = self.token.character_name

        mock_filter = MagicMock()
        mock_filter.first.return_value = mock_character
        mock_select_related.return_value.filter.return_value = mock_filter
        # when
        result = update_character(self.token.character_id)
        # then
        self.assertTrue(result)

    @patch(MODULE_PATH + ".update_character_mining")
    def test_update_character_mining(self, mock_char_mining):
        # given
        expected_return_value = f"Finished Mining for: {self.token.character_name}"
        mock_char_mining.return_value = expected_return_value
        # when
        result = update_char_mining_ledger(self.token.character_id)
        # then
        self.assertTrue(mock_char_mining.called)
        self.assertEqual(expected_return_value, result)

    @patch(MODULE_PATH + ".update_character_wallet")
    def test_update_character_wallet(self, mock_char_wallet):
        # given
        expected_return_value = (
            f"Finished wallet transactions for: {self.token.character_name}"
        )
        mock_char_wallet.return_value = expected_return_value
        # when
        result = update_char_wallet(self.token.character_id)
        # then
        self.assertTrue(mock_char_wallet.called)
        self.assertEqual(expected_return_value, result)

    @patch(MODULE_PATH + ".update_char_mining_ledger.si")
    @patch(MODULE_PATH + ".update_char_wallet.si")
    @patch(MODULE_PATH + ".Token.get_token")
    @patch(MODULE_PATH + ".EveCharacter.objects.get_character_by_id")
    @patch(MODULE_PATH + ".CharacterAudit.objects.update_or_create")
    @patch(MODULE_PATH + ".CharacterAudit.objects.select_related")
    def test_update_character_from_token(
        self,
        mock_check_char,
        mock_update_or_create,
        mock_get_character_by_id,
        mock_get_token,
        mock_char_wallet,
        mock_char_mining,
    ):
        # given
        mock_check_char.return_value.filter.return_value.first.return_value = None
        mock_token = MagicMock()
        mock_token.valid_access_token.return_value = True
        mock_get_token.return_value = mock_token
        mock_character = MagicMock()
        mock_character.last_update_mining = timezone.now() - timedelta(days=1)
        mock_character.last_update_wallet = timezone.now() - timedelta(days=1)
        mock_character.last_update_planetary = timezone.now() - timedelta(days=1)
        mock_update_or_create.return_value = (mock_character, True)
        # when
        update_character(self.token.character_id)
        # then
        mock_get_token.assert_called_once_with(
            self.token.character_id, CharacterAudit.get_esi_scopes()
        )
        mock_token.valid_access_token.assert_called_once()
        mock_get_character_by_id.assert_called_once_with(mock_token.character_id)
        mock_update_or_create.assert_called_once()
        self.assertTrue(mock_char_wallet.called)
        self.assertTrue(mock_char_mining.called)

    @patch(MODULE_PATH + ".Token.get_token")
    @patch(MODULE_PATH + ".CharacterAudit.objects.update_or_create")
    @patch(MODULE_PATH + ".CharacterAudit.objects.select_related")
    def test_update_character_token_expired_token(
        self, mock_check_char, mock_update_or_create, mock_get_token
    ):
        # given
        mock_check_char.return_value.filter.return_value.first.return_value = None
        mock_token = MagicMock()
        mock_token.valid_access_token.side_effect = TokenExpiredError
        mock_get_token.return_value = mock_token
        mock_character = MagicMock()
        mock_update_or_create.return_value = (mock_character, True)
        # when
        result = update_character(self.token.character_id)
        # then
        self.assertFalse(result)
        mock_get_token.assert_called_once_with(
            self.token.character_id, CharacterAudit.get_esi_scopes()
        )
        mock_token.valid_access_token.assert_called_once()

    @patch(MODULE_PATH + ".Token.get_token")
    @patch(MODULE_PATH + ".CharacterAudit.objects.select_related")
    def test_update_character_token_no_access_token(
        self, mock_check_char, mock_get_token
    ):
        # given
        mock_check_char.return_value.filter.return_value.first.return_value = None
        mock_token = MagicMock()
        mock_token.valid_access_token.return_value = False
        mock_get_token.return_value = mock_token
        # when
        result = update_character(self.token.character_id)
        # then
        mock_token.valid_access_token.assert_called_once()
        self.assertFalse(result)

    @patch(MODULE_PATH + ".Token.get_token")
    @patch(MODULE_PATH + ".CharacterAudit.objects.select_related")
    @patch(MODULE_PATH + ".logger")
    def test_update_character_token_no_token(
        self, mock_logger, mock_check_char, mock_get_token
    ):
        # given
        mock_check_char.return_value.filter.return_value.first.return_value = None
        mock_get_token.return_value = False
        # when
        result = update_character(self.token.character_id)
        # then
        self.assertFalse(result)
        mock_logger.info.assert_called_once_with("No Tokens for %s", 1001)

    @patch(MODULE_PATH + ".update_corp.apply_async")
    @patch(MODULE_PATH + ".logger")
    def test_update_all_cors(self, mock_logger, mock_update_character):
        # given
        corporation_count = CorporationAudit.objects.count()
        # when
        update_all_corps()
        # then
        self.assertEqual(mock_update_character.call_count, corporation_count)
        mock_logger.info.assert_called_once_with(
            "Queued %s Corp Audit Updates", corporation_count
        )

    @patch(MODULE_PATH + ".update_corp_wallet.si")
    def test_update_corp(self, mock_corp_wallet):
        # when
        update_corp(self.corporation.corporation_id)
        # then
        self.assertTrue(mock_corp_wallet.called)

    @patch(MODULE_PATH + ".update_corp_wallet_division")
    def test_update_corp_wallet(self, mock_corp_wallet_division):
        # given
        expected_return_value = (
            f"Finished wallet divs for: {self.corporation.corporation_name}"
        )
        mock_corp_wallet_division.return_value = expected_return_value
        # when
        result = update_corp_wallet(self.corporation.corporation_id)
        # then
        self.assertTrue(mock_corp_wallet_division.called)
        self.assertEqual(expected_return_value, result)

    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects.all")
    @patch(MODULE_PATH + ".CharacterOwnership.objects.get")
    @patch(MODULE_PATH + ".notify")
    def test_no_expired_alarms(self, mock_notify, mock_get_ownership, mock_all_planets):
        # Setup mock return values
        mock_all_planets.return_value = []

        # Call the function
        check_planetary_alarms()

        # Check that notify was not called
        mock_notify.assert_not_called()

    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects.all")
    @patch(MODULE_PATH + ".CharacterOwnership.objects.get")
    @patch(MODULE_PATH + ".notify")
    def test_expired_alarms_notification_sent(
        self, mock_notify, mock_get_ownership, mock_all_planets
    ):
        # Setup mock return values
        planet = MagicMock()
        planet.is_expired.return_value = True
        planet.notification_sent = False
        planet.notification = True
        planet.planet.character.character.character_id = 1999
        planet.planet.character.character.character_name = "Test Character"
        planet.planet.planet.name = "Test Planet"
        mock_all_planets.return_value = [planet]

        owner = MagicMock()
        owner.user.profile.main_character.character_id = 1999
        owner.user.profile.main_character.character_ownership.user.character_ownerships.all.return_value.values_list.return_value = [
            1999
        ]
        mock_get_ownership.return_value = owner

        # Call the function
        check_planetary_alarms()

        # Check that notify was called
        mock_notify.assert_called_once()
        planet.save.assert_called_once()

    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects.all")
    @patch(MODULE_PATH + ".CharacterOwnership.objects.get")
    @patch(MODULE_PATH + ".notify")
    def test_expired_alarms_notification_already_sent(
        self, mock_notify, mock_get_ownership, mock_all_planets
    ):
        # Setup mock return values
        planet = MagicMock()
        planet.is_expired.return_value = True
        planet.notification_sent = True
        planet.notification = True
        mock_all_planets.return_value = [planet]

        # Call the function
        check_planetary_alarms()

        # Check that notify was not called
        mock_notify.assert_not_called()
        planet.save.assert_not_called()

    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects.all")
    @patch(MODULE_PATH + ".CharacterOwnership.objects.get")
    @patch(MODULE_PATH + ".notify")
    def test_notification_message_format(
        self, mock_notify, mock_get_ownership, mock_all_planets
    ):
        # Setup mock return values
        planet = MagicMock()
        planet.is_expired.return_value = True
        planet.notification_sent = False
        planet.notification = True
        planet.planet.character.character.character_id = 1999
        planet.planet.character.character.character_name = "Test Character"
        planet.planet.planet.name = "Test Planet"
        mock_all_planets.return_value = [planet]

        owner = MagicMock()
        owner.user.profile.main_character.character_id = 1999
        owner.user.profile.main_character.character_ownership.user.character_ownerships.all.return_value.values_list.return_value = [
            1999
        ]
        mock_get_ownership.return_value = owner

        # Call the function
        check_planetary_alarms()

        # Check that notify was called with correct message
        expected_message = format_html(
            "Following Planet Extractor Heads have expired: \n{}",
            "Test Character on Test Planet",
        )
        mock_notify.assert_called_once_with(
            title=_("Planetary Extractor Heads Expired"),
            message=expected_message,
            user=owner.user,
            level="warning",
        )
        planet.save.assert_called_once()

    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects.all")
    @patch(MODULE_PATH + ".CharacterOwnership.objects.get")
    @patch(MODULE_PATH + ".notify")
    def test_main_id_set(self, mock_notify, mock_get_ownership, mock_all_planets):
        # Setup mock return values
        planet = MagicMock()
        planet.is_expired.return_value = True
        planet.notification_sent = False
        planet.notification = True
        planet.planet.character.character.character_id = 1999
        planet.planet.character.character.character_name = "Test Character"
        planet.planet.planet.name = "Test Planet"
        mock_all_planets.return_value = [planet]

        owner = MagicMock()
        owner.user.profile.main_character.character_id = 1999
        owner.user.profile.main_character.character_ownership.user.character_ownerships.all.return_value.values_list.return_value = [
            1999
        ]
        mock_get_ownership.return_value = owner

        # Call the function
        check_planetary_alarms()

        # Check that notify was called with correct message
        expected_message = format_html(
            "Following Planet Extractor Heads have expired: \n{}",
            "Test Character on Test Planet",
        )
        mock_notify.assert_called_once_with(
            title=_("Planetary Extractor Heads Expired"),
            message=expected_message,
            user=owner.user,
            level="warning",
        )
        planet.save.assert_called_once()

        # Verify that main_id was set correctly
        self.assertEqual(owner.user.profile.main_character.character_id, 1999)

    @patch(MODULE_PATH + ".update_character_planetary")
    def test_update_char_planets(self, mock_char_planet):
        # given
        expected_return_value = ("Finished planets update for: %s", "Gneuten")
        mock_char_planet.return_value = expected_return_value
        # when
        result = update_char_planets(self.token.character_id)
        # then
        self.assertTrue(mock_char_planet.called)
        self.assertEqual(expected_return_value, result)

    @patch(MODULE_PATH + ".update_character_planetary_details")
    def test_update_char_planets_details(self, mock_char_planetdetails):
        # given
        expected_return_value = ("Finished planets data update for: %s", "Gneuten")
        mock_char_planetdetails.return_value = expected_return_value
        # when
        result = update_char_planets_details(1001, 4001)
        # then
        self.assertTrue(mock_char_planetdetails.called)
        self.assertEqual(expected_return_value, result)

    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects.all")
    @patch(MODULE_PATH + ".CharacterOwnership.objects.get")
    @patch(MODULE_PATH + ".notify")
    def test_check_planetary_alarms_many(
        self, mock_notify, mock_get_ownership, mock_all_planets
    ):
        # Setup mock return values
        planet = MagicMock()
        planet.is_expired.return_value = True
        planet.notification_sent = False
        planet.notification = True
        planet.planet.character.character.character_id = 1001
        planet.planet.character.character.character_name = "Gneuten"
        planet.planet.planet.name = "Test Planet I"

        planet2 = MagicMock()
        planet2.is_expired.return_value = True
        planet2.notification_sent = False
        planet2.notification = True
        planet2.planet.character.character.character_id = 1001
        planet2.planet.character.character.character_name = "Gneuten"
        planet2.planet.planet.name = "Test Planet II"

        mock_all_planets.return_value = [planet, planet2]

        mock_get_ownership.return_value = self.character_ownership

        # Call the function
        check_planetary_alarms()

        # Check that notify was called with correct message
        expected_message = format_html(
            "Following Planet Extractor Heads have expired: \n{}\n{}",
            "Gneuten on Test Planet I",
            "Gneuten on Test Planet II",
        )
        mock_notify.assert_called_once_with(
            title=_("Planetary Extractor Heads Expired"),
            message=expected_message,
            user=self.user,
            level="warning",
        )
        planet.save.assert_called_once()

    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects.all")
    @patch(MODULE_PATH + ".CharacterOwnership.objects.get")
    @patch(MODULE_PATH + ".notify")
    def test_character_ownership_does_not_exist(
        self, mock_notify, mock_get_ownership, mock_all_planets
    ):
        # Setup mock return values
        planet = MagicMock()
        planet.is_expired.return_value = True
        planet.notification_sent = False
        planet.notification = True
        planet.planet.character.character.character_id = 1999
        mock_all_planets.return_value = [planet]

        mock_get_ownership.side_effect = CharacterOwnership.DoesNotExist

        # Call the function
        check_planetary_alarms()

        # Check that notify was not called
        mock_notify.assert_not_called()
        planet.save.assert_not_called()

    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects.all")
    @patch(MODULE_PATH + ".CharacterOwnership.objects.get")
    @patch(MODULE_PATH + ".notify")
    def test_attribute_error(self, mock_notify, mock_get_ownership, mock_all_planets):
        # Setup mock return values
        planet = MagicMock()
        planet.is_expired.return_value = True
        planet.notification_sent = False
        planet.notification = True
        planet.planet.character.character.character_id = 1999
        mock_all_planets.return_value = [planet]

        mock_get_ownership.side_effect = AttributeError

        # Call the function
        check_planetary_alarms()

        # Check that notify was not called
        mock_notify.assert_not_called()
        planet.save.assert_not_called()

    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects.all")
    @patch(MODULE_PATH + ".CharacterOwnership.objects.get")
    @patch(MODULE_PATH + ".notify")
    def test_character_id_not_in_alts(
        self, mock_notify, mock_get_ownership, mock_all_planets
    ):
        # Setup mock return values
        planet = MagicMock()
        planet.is_expired.return_value = True
        planet.notification_sent = False
        planet.notification = True
        planet.planet.character.character.character_id = 1004
        planet.planet.character.character.character_name = "Test Character"
        planet.planet.planet.name = "Test Planet I"
        mock_all_planets.return_value = [planet]

        # Setup mock return values
        planet2 = MagicMock()
        planet2.is_expired.return_value = True
        planet2.notification_sent = False
        planet2.notification = True
        planet2.planet.character.character.character_id = 1001
        planet2.planet.character.character.character_name = "Gneuten"
        planet2.planet.planet.name = "Test Planet II"

        # Setup mock return values
        planet3 = MagicMock()
        planet3.is_expired.return_value = True
        planet3.notification_sent = False
        planet3.notification = True
        planet3.planet.character.character.character_id = 1005
        planet3.planet.character.character.character_name = "Gorthd"
        planet3.planet.planet.name = "Test Planet III"

        mock_all_planets.return_value = [planet, planet2, planet3]

        owner = MagicMock()
        owner.user.profile.main_character.character_id = 1001
        owner.user.profile.main_character.character_ownership.user.character_ownerships.all.return_value.values_list.return_value = [
            1001,
            1003,
        ]
        mock_get_ownership.return_value = owner

        # Call the function
        check_planetary_alarms()

        planet.save.assert_called_once()

        # Verify that main_id was set correctly
        self.assertEqual(owner.user.profile.main_character.character_id, 1001)

    @patch(MODULE_PATH + ".CharacterPlanetDetails.objects.all")
    @patch(MODULE_PATH + ".CharacterOwnership.objects.get")
    @patch(MODULE_PATH + ".notify")
    def test_character_id_in_alts(
        self, mock_notify, mock_get_ownership, mock_all_planets
    ):
        add_character_to_user(self.user, EveCharacter.objects.get(character_id=1002))
        # Setup mock return values
        planet = MagicMock()
        planet.is_expired.return_value = True
        planet.notification_sent = False
        planet.notification = True
        planet.planet.character.character.character_id = 1002
        planet.planet.character.character.character_name = "Gneuten"
        planet.planet.planet.name = "Test Planet I"
        mock_all_planets.return_value = [planet]

        owner = MagicMock()
        owner.user.profile.main_character.character_id = 1001
        owner.user.profile.main_character.character_ownership.user.character_ownerships.all.return_value.values_list.return_value = [
            1001,
            1002,
        ]
        mock_get_ownership.return_value = owner

        # Call the function
        check_planetary_alarms()

        # Check that notify was called with correct message
        expected_message = format_html(
            "Following Planet Extractor Heads have expired: \n{}",
            "Gneuten on Test Planet I",
        )
        mock_notify.assert_called_once_with(
            title=_("Planetary Extractor Heads Expired"),
            message=expected_message,
            user=owner.user,
            level="warning",
        )
        planet.save.assert_called_once()

        # Verify that main_id was set correctly
        self.assertEqual(owner.user.profile.main_character.character_id, 1001)



from unittest.mock import MagicMock, call, patch

from django.db import IntegrityError
from django.test import TestCase
from django.utils.translation import gettext_lazy as _

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import create_user_from_evecharacter

from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all
from ledger.tests.testdata.load_planetary import load_planetary

MODULE_PATH = "ledger.tasks"


class TestTasks(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()
        load_planetary()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
            ],
        )

        cls.user2, cls.character_ownership2 = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
            ],
        )

        cls.token = cls.user.token_set.first()
        cls.corporation = cls.character_ownership.character.corporation



import json
from collections import defaultdict
from decimal import Decimal

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import Case, DecimalField, F, Q, Sum, Value, When
from django.db.models.functions import Coalesce, Round
from django.utils import timezone

from allianceauth.authentication.models import UserProfile

from ledger import app_settings
from ledger.hooks import get_extension_logger
from ledger.models.general import EveEntity
from ledger.view_helpers.core import events_filter

logger = get_extension_logger(__name__)

# Ref Types PvE - Tax Income
BOUNTY_PRIZES = ["bounty_prizes"]
ESS_TRANSFER = ["ess_escrow_transfer"]
INCURSION = ["corporate_reward_payout"]
MISSION = ["agent_mission_reward", "agent_mission_time_bonus_reward"]
DAILY_GOAL_REWARD = ["daily_goal_payouts"]
CITADEL_INCOME = ["industry_job_tax", "reprocessing_tax"]

# Filters
BOUNTY_FILTER = Q(ref_type__in=BOUNTY_PRIZES, amount__gt=0)
ESS_FILTER = Q(ref_type__in=ESS_TRANSFER, amount__gt=0)
INCURSION_FILTER = Q(ref_type__in=INCURSION, amount__gt=0)
MISSION_FILTER = Q(ref_type__in=MISSION, amount__gt=0)
DAILY_GOAL_REWARD_FILTER = Q(ref_type__in=DAILY_GOAL_REWARD, amount__gt=0)
CITADEL_FILTER = Q(ref_type__in=CITADEL_INCOME, amount__gt=0)

MISC_FILTER = (
    INCURSION_FILTER | MISSION_FILTER | DAILY_GOAL_REWARD_FILTER | CITADEL_FILTER
)


class CorpWalletQueryFilter(models.QuerySet):
    def _convert_corp_tax(self, ess: models.QuerySet) -> Decimal:
        """Convert corp tax to correct amount for character ledger"""
        amount = (ess / app_settings.LEDGER_CORP_TAX) * (
            100 - app_settings.LEDGER_CORP_TAX
        )
        return amount

    def annotate_bounty_income(self) -> models.QuerySet:
        return self.annotate(
            bounty_income=Coalesce(
                Sum(
                    "amount",
                    filter=(BOUNTY_FILTER),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_ess_income(self, is_character_ledger: bool = False) -> models.QuerySet:
        if is_character_ledger:
            return self.annotate(
                ess_income=Round(
                    Coalesce(
                        Sum(
                            self._convert_corp_tax(F("amount")),
                            filter=(ESS_FILTER),
                        ),
                        Value(0),
                        output_field=DecimalField(),
                    ),
                    precision=2,
                )
            )
        return self.annotate(
            ess_income=Coalesce(
                Sum(
                    F("amount"),
                    filter=(ESS_FILTER),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    # pylint: disable=duplicate-code
    def annotate_mission_income(self) -> models.QuerySet:
        return self.annotate(
            mission_income=Coalesce(
                Sum(
                    "amount",
                    filter=(MISSION_FILTER),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    # pylint: disable=duplicate-code
    def annotate_incursion_income(self) -> models.QuerySet:
        return self.annotate(
            incursion_income=Coalesce(
                Sum(
                    "amount",
                    filter=(INCURSION_FILTER),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_daily_goal_income(
        self, is_character_ledger: bool = False
    ) -> models.QuerySet:
        if is_character_ledger:
            return self.annotate(
                daily_goal_income=Round(
                    Coalesce(
                        Sum(
                            self._convert_corp_tax(F("amount")),
                            filter=(DAILY_GOAL_REWARD_FILTER),
                        ),
                        Value(0),
                        output_field=DecimalField(),
                    ),
                    precision=2,
                )
            )
        return self.annotate(
            daily_goal_income=Coalesce(
                Sum(
                    F("amount"),
                    filter=(DAILY_GOAL_REWARD_FILTER),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_citadel_income(self) -> models.QuerySet:
        return self.annotate(
            citadel_income=Coalesce(
                Sum(
                    "amount",
                    filter=(CITADEL_FILTER),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    # pylint: disable=duplicate-code
    def annotate_miscellaneous(self) -> models.QuerySet:
        return self.annotate(
            miscellaneous=Coalesce(
                Sum(
                    "amount",
                    filter=(MISC_FILTER),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )


class CorpWalletQuerySet(CorpWalletQueryFilter):
    # pylint: disable=unused-argument
    def _get_linked_chars(self, entity_ids: list, corporations=None) -> tuple:
        main_and_alts = {}
        entity_list = []
        # ESS, Daily, Bounty Payout Corps
        eve_npc_entities = [1000125, 1000132, 1000413]

        accounts = UserProfile.objects.filter(
            main_character__isnull=False,
        ).select_related(
            "user__profile__main_character",
            "main_character__character_ownership",
            "main_character__character_ownership__user__profile",
            "main_character__character_ownership__user__profile__main_character",
        )

        for account in accounts:
            alts = account.user.character_ownerships.all().values_list(
                "character__character_id", flat=True
            )
            main = account.main_character

            try:
                main_char = EveEntity.objects.get(eve_id=main.character_id)
            except ObjectDoesNotExist:
                main_char = EveEntity(
                    eve_id=main.character_id,
                    name=main.character_name,
                    category="character",
                )

            main_and_alts[main_char] = alts
            entity_list.extend(alts)

        missing_entitys = set(entity_ids) - set(entity_list)

        entitys = EveEntity.objects.filter(eve_id__in=missing_entitys)
        # Get All Eve Entities
        for entity in entitys:
            # Ensure we don't include NPC entities
            if entity.eve_id in eve_npc_entities:
                continue
            try:
                entity_id = entity.eve_id
                main_and_alts[entity] = [entity_id]
                entity_list.append(entity_id)
            except AttributeError:
                continue

        return main_and_alts, set(entity_list)

    def annotate_ledger_data(self, queryset) -> models.QuerySet:
        """Annotate the queryset with ledger data"""
        return (
            queryset.annotate_bounty_income()
            .annotate_ess_income()
            .annotate_mission_income()
            .annotate_incursion_income()
            .annotate_daily_goal_income()
            .annotate_citadel_income()
            .annotate_miscellaneous()
        )

    def generate_ledger(self, corporations: list) -> models.QuerySet:
        """Generate the ledger for the corporation"""
        # Filter Corporations6
        qs = self.filter(
            Q(division__corporation__corporation__corporation_id__in=corporations)
        )
        # Get all Character IDs
        entity_ids_second = self.values_list("second_party_id", flat=True)
        entity_ids_first = self.values_list("first_party_id", flat=True)
        entity_ids = set(entity_ids_first) | set(entity_ids_second)

        # Get all linked Characters
        main_and_alts, chars_list = self._get_linked_chars(entity_ids, corporations)

        # Filter queryset with the linked characters
        qs = qs.filter(
            Q(first_party_id__in=chars_list) | Q(second_party_id__in=chars_list)
        )
        # Exclude Tax Events
        qs = events_filter(qs)

        # Create annotation cases
        main_entity_id_cases = []
        main_entity_name_cases = []
        alts_cases = []

        # Create the cases for the annotations
        for main, alts in main_and_alts.items():
            # Filter alts to include only those present in entity_ids
            filtered_alts = [alt for alt in alts if alt in entity_ids]
            if filtered_alts:
                main_entity_id_cases.append(
                    When(
                        Q(second_party_id__in=filtered_alts)
                        | Q(first_party_id__in=filtered_alts),
                        then=Value(main.eve_id),
                    )
                )
                main_entity_name_cases.append(
                    When(
                        Q(second_party_id__in=filtered_alts)
                        | Q(first_party_id__in=filtered_alts),
                        then=Value(main.name),
                    )
                )
                alts_cases.append(
                    When(
                        Q(second_party_id__in=filtered_alts)
                        | Q(first_party_id__in=filtered_alts),
                        then=Value(json.dumps(filtered_alts)),
                    )
                )

        # Annotate the queryset
        qs = (
            qs.annotate(
                main_entity_id=Case(
                    *main_entity_id_cases,
                    output_field=models.IntegerField(),
                ),
                main_entity_name=Case(
                    *main_entity_name_cases,
                    output_field=models.CharField(),
                ),
                alts=Case(
                    *alts_cases,
                    default=Value("[]"),
                    output_field=models.JSONField(),
                ),
            )
            .values(
                "main_entity_id",
                "main_entity_name",
                "alts",
            )
            .distinct()
        )

        return qs

    # pylint: disable=duplicate-code, too-many-positional-arguments
    def generate_template(
        self,
        amounts: defaultdict,
        filter_date: timezone.datetime,
        character_ids: list,
        corporations_ids: list,
        entity_type=None,
    ) -> dict:
        """Generate data template for the ledger character information view"""
        # Define the type names
        type_names = [
            "ess_income",
            "daily_goal_income",
        ]

        if entity_type == "corporation":
            type_names += [
                "bounty_income",
                "mission_income",
                "citadel_income",
                "incursion_income",
            ]
        qs = self

        # Filter Corporations
        qs = self.filter(
            Q(division__corporation__corporation__corporation_id__in=corporations_ids)
        )
        # Filter Characters
        qs = qs.filter(
            Q(first_party_id__in=character_ids) | Q(second_party_id__in=character_ids)
        )

        # Exclude Corp Tax Events
        qs = events_filter(qs)

        # Annotate the queryset
        qs = self.annotate_ledger_data(qs)

        annotations = {}
        for type_name in type_names:
            annotations[f"{type_name}_total_amount"] = Coalesce(
                Sum(
                    Case(
                        When(
                            **{
                                f"{type_name}__isnull": False,
                                "date__year": filter_date.year,
                            },
                            then=F(type_name),
                        )
                    )
                ),
                Value(0),
                output_field=DecimalField(),
            )
            annotations[f"{type_name}_total_amount_day"] = Coalesce(
                Sum(
                    Case(
                        When(
                            **{
                                f"{type_name}__isnull": False,
                                "date__year": filter_date.year,
                                "date__month": filter_date.month,
                                "date__day": filter_date.day,
                            },
                            then=F(type_name),
                        )
                    )
                ),
                Value(0),
                output_field=DecimalField(),
            )
            annotations[f"{type_name}_total_amount_hour"] = Coalesce(
                Sum(
                    Case(
                        When(
                            **{
                                f"{type_name}__isnull": False,
                                "date__year": filter_date.year,
                                "date__month": filter_date.month,
                                "date__day": filter_date.day,
                                "date__hour": filter_date.hour,
                            },
                            then=F(type_name),
                        )
                    )
                ),
                Value(0),
                output_field=DecimalField(),
            )

        qs = qs.aggregate(**annotations)

        # Assign the results to the amounts dictionary
        for type_name in type_names:
            if entity_type != "corporation":
                amounts[type_name]["total_amount"] = self._convert_corp_tax(
                    qs[f"{type_name}_total_amount"]
                )
                amounts[type_name]["total_amount_day"] = self._convert_corp_tax(
                    qs[f"{type_name}_total_amount_day"]
                )
                amounts[type_name]["total_amount_hour"] = self._convert_corp_tax(
                    qs[f"{type_name}_total_amount_hour"]
                )
            else:
                amounts[type_name]["total_amount"] = qs[f"{type_name}_total_amount"]
                amounts[type_name]["total_amount_day"] = qs[
                    f"{type_name}_total_amount_day"
                ]
                amounts[type_name]["total_amount_hour"] = qs[
                    f"{type_name}_total_amount_hour"
                ]

        return amounts

    def annotate_billboard(self, chars: list) -> models.QuerySet:
        qs = self.filter(Q(first_party_id__in=chars) | Q(second_party_id__in=chars))
        # Exclude Corp Tax Events
        qs = events_filter(qs)
        return qs


class CorpWalletManagerBase(models.Manager):
    pass


CorpWalletManager = CorpWalletManagerBase.from_queryset(CorpWalletQuerySet)



from decimal import Decimal
from unittest.mock import MagicMock, patch

from django.test import TestCase
from esi.models import Token

from allianceauth.corputils.models import CorpMember, CorpStats
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import add_character_to_user, create_user_from_evecharacter

from ledger.managers.corpjournal_manager import CorpWalletManager
from ledger.models import CorporationWalletDivision, CorporationWalletJournalEntry
from ledger.models.general import EveEntity
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all

MODULE_PATH = "ledger.managers.corpjournal_manager"


class CharManagerQuerySetTest(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()
        cls.manager = CorpWalletManager()
        cls.corp = EveCorporationInfo.objects.get(corporation_id=2001)

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.advanced_access",
            ],
        )
        cls.user2, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
                "ledger.advanced_access",
            ],
        )

    def test_annotate_bounty(self):
        qs = CorporationWalletJournalEntry.objects.annotate_bounty_income()
        self.assertIsNotNone(qs)
        self.assertIn("bounty_income", qs.query.annotations)

    def test_annotate_ess(self):
        qs = CorporationWalletJournalEntry.objects.annotate_ess_income()
        self.assertIsNotNone(qs)
        self.assertIn("ess_income", qs.query.annotations)

    def test_annotate_mission(self):
        qs = CorporationWalletJournalEntry.objects.annotate_mission_income()
        self.assertIsNotNone(qs)
        self.assertIn("mission_income", qs.query.annotations)

    def test_annotate_daily_goal(self):
        qs = CorporationWalletJournalEntry.objects.annotate_daily_goal_income()
        self.assertIsNotNone(qs)
        self.assertIn("daily_goal_income", qs.query.annotations)

    def test_annotate_ledger(self):
        add_character_to_user(self.user, EveCharacter.objects.get(character_id=1002))
        add_character_to_user(self.user, EveCharacter.objects.get(character_id=1004))
        add_character_to_user(self.user, EveCharacter.objects.get(character_id=1003))
        add_character_to_user(self.user, EveCharacter.objects.get(character_id=1006))

        test9999 = EveEntity.objects.create(
            eve_id=9999,
            category=EveEntity.CATEGORY_CHARACTER,
            name="Test9999",
        )

        test9998 = EveEntity.objects.create(
            eve_id=9998,
            category=EveEntity.CATEGORY_CHARACTER,
            name="Test9999",
        )

        # Create CorporationWalletJournalEntry objects
        CorporationWalletJournalEntry.objects.create(
            division=CorporationWalletDivision.objects.get(id=1),
            amount=100_000,
            balance=100_000_000,
            context_id=0,
            context_id_type="division",
            date="2024-03-19T14:00:00Z",
            description="Test",
            first_party=EveEntity.objects.get(eve_id=1000125),
            entry_id=298,
            reason="",
            ref_type="bounty_prizes",
            second_party=test9998,
            tax=0,
            tax_receiver_id=0,
        )
        CorporationWalletJournalEntry.objects.create(
            division=CorporationWalletDivision.objects.get(id=1),
            amount=100_000,
            balance=100_000_000,
            context_id=0,
            context_id_type="division",
            date="2024-03-19T14:00:00Z",
            description="Test",
            first_party=EveEntity.objects.get(eve_id=1000125),
            entry_id=299,
            reason="",
            ref_type="bounty_prizes",
            second_party=test9999,
            tax=0,
            tax_receiver_id=0,
        )

        self.client.force_login(self.user)
        result = CorporationWalletJournalEntry.objects.all().generate_ledger([2001])

        expected_result = [
            {
                "main_entity_id": 1001,
                "main_entity_name": "Gneuten",
                "alts": [1001, 1002],
            },
            {
                "main_entity_id": 1010,
                "main_entity_name": "Test1",
                "alts": [1010],
            },
            {
                "main_entity_id": 1011,
                "main_entity_name": "Test2",
                "alts": [1011],
            },
            {
                "main_entity_id": 1012,
                "main_entity_name": "Test3",
                "alts": [1012],
            },
            {
                "main_entity_id": 1013,
                "main_entity_name": "Test4",
                "alts": [1013],
            },
            {
                "main_entity_id": 1014,
                "main_entity_name": "Test5",
                "alts": [1014],
            },
            {
                "main_entity_id": 1015,
                "main_entity_name": "Test6",
                "alts": [1015],
            },
            {
                "main_entity_id": 1016,
                "main_entity_name": "Test7",
                "alts": [1016],
            },
            {
                "main_entity_id": 1017,
                "main_entity_name": "Test8",
                "alts": [1017],
            },
            {
                "main_entity_id": 1018,
                "main_entity_name": "Test9",
                "alts": [1018],
            },
            {
                "main_entity_id": 1019,
                "main_entity_name": "Test10",
                "alts": [1019],
            },
            {
                "main_entity_id": 1020,
                "main_entity_name": "Test11",
                "alts": [1020],
            },
            {
                "main_entity_id": 1021,
                "main_entity_name": "Test12",
                "alts": [1021],
            },
            {
                "main_entity_id": 9998,
                "main_entity_name": "Test9999",
                "alts": [9998],
            },
            {
                "main_entity_id": 9999,
                "main_entity_name": "Test9999",
                "alts": [9999],
            },
        ]

        sorted_result = sorted(list(result), key=lambda x: x["main_entity_id"])
        sorted_expected_result = sorted(
            expected_result, key=lambda x: x["main_entity_id"]
        )

        self.assertEqual(sorted_result, sorted_expected_result)

    def test_annotate_ledger_no_data(self):
        result = CorporationWalletJournalEntry.objects.all().generate_ledger([99999])

        self.assertEqual(list(result), [])

    def test_annotate_ledger_no_data_no_corporations(self):
        result = CorporationWalletJournalEntry.objects.all().generate_ledger([])

        self.assertEqual(list(result), [])



"""PvE Views"""

from datetime import datetime

# Django
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import render

from ledger.hooks import get_extension_logger

# Ledger
from ledger.view_helpers.core import add_info_to_context

logger = get_extension_logger(__name__)


@login_required
@permission_required("ledger.basic_access", "ledger.advanced_access")
def corporation_ledger(request, corporation_pk):
    """
    Corporation Ledger
    """
    # pylint: disable=duplicate-code
    current_year = datetime.now().year
    years = [current_year - i for i in range(6)]

    context = {
        "years": years,
        "entity_pk": corporation_pk,
        "entity_type": "corporation",
    }
    context = add_info_to_context(request, context)
    return render(request, "ledger/corpledger/corporation_ledger.html", context=context)


@login_required
@permission_required("ledger.basic_access", "ledger.advanced_access")
def corporation_admin(request):
    """
    Corporation Admin
    """
    context = {}
    context = add_info_to_context(request, context)
    return render(
        request, "ledger/corpledger/admin/corporation_admin.html", context=context
    )
